import json
import unittest
from aiohttp import ClientResponseError
from aioresponses import aioresponses, CallbackResult

from eyepop import EyePopSdk
from tests.data.base_endpoint_test import BaseEndpointTest


class TestEndpointConnect(BaseEndpointTest):

    def test_missing_secret(self):
        with self.assertRaises(Exception):
            EyePopSdk.dataEndpoint()

    @aioresponses()
    def test_connect_ok(self, mock: aioresponses):

        self.setup_base_mock(mock)

        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        endpoint = EyePopSdk.dataEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                          account_id=self.test_eyepop_account_id)
        try:
            endpoint.connect()
        finally:
            endpoint.disconnect()

        self.assertBaseMock(mock)

    @aioresponses()
    def test_connect_unauthorized(self, mock: aioresponses):
        self.setup_base_mock(mock)

        mock.post(f'{self.test_eyepop_url}/authentication/token', status=401, body="test unauthorized")
        # automatic call to get pop comp and store
        endpoint = EyePopSdk.dataEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                          account_id=self.test_eyepop_account_id)
        with self.assertRaises(ClientResponseError) as context:
            try:
                endpoint.connect()
            finally:
                endpoint.disconnect()
        self.assertEqual(context.exception.status, 401)

    @aioresponses()
    def test_connect_reauth_expired(self, mock: aioresponses):

        self.setup_base_mock(mock)

        auth_calls = 0

        def first_auth(url, **kwargs) -> CallbackResult:
            nonlocal auth_calls
            auth_calls += 1
            return CallbackResult(status=200, body=json.dumps(
                {'expires_in': 1, 'token_type': 'Bearer', 'access_token': self.test_access_token}))

        def second_auth(url, **kwargs) -> CallbackResult:
            nonlocal auth_calls
            auth_calls += 1
            return CallbackResult(status=200, body=json.dumps(
                {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))

        mock.post(f'{self.test_eyepop_url}/authentication/token', callback=first_auth)
        mock.post(f'{self.test_eyepop_url}/authentication/token', callback=second_auth)
        endpoint = EyePopSdk.dataEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                          account_id=self.test_eyepop_account_id)
        try:
            endpoint.connect()
        finally:
            endpoint.disconnect()

        self.assertBaseMock(mock)

        self.assertEqual(auth_calls, 2)

    @aioresponses()
    def test_connect_reauth_401(self, mock: aioresponses):
        self.setup_base_mock(mock)

        auth_calls = 0

        def auth(url, **kwargs) -> CallbackResult:
            nonlocal auth_calls
            if auth_calls == 0:
                result = CallbackResult(status=200, body=json.dumps(
                    {'expires_in': 1000, 'token_type': 'Bearer', 'access_token': self.test_expired_access_token}))
            else:
                result = CallbackResult(status=200, body=json.dumps(
                    {'expires_in': 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
            auth_calls += 1
            return result

        mock.post(f'{self.test_eyepop_url}/authentication/token', callback=auth, repeat=True)
        endpoint = EyePopSdk.dataEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                          account_id=self.test_eyepop_account_id)
        try:
            endpoint.connect()
        finally:
            endpoint.disconnect()

        self.assertBaseMock(mock)
        self.assertEqual(auth_calls, 2)


if __name__ == '__main__':
    unittest.main()



import pyarrow as pa

from eyepop.data.data_types import MIME_TYPE_APACHE_ARROW_FILE

""" Arrow schema for Asset export/import form Data API. """

MIME_TYPE_APACHE_ARROW_FILE_VERSIONED = f"{MIME_TYPE_APACHE_ARROW_FILE};version=0.0"

_object_fields = [
    # Python type is string, there should be only as small number of
    # classLabel values per prediction, hence the dictionary encoding.
    pa.field("classLabel", pa.dictionary(pa.int32(), pa.string())),
    pa.field("confidence", pa.float16()),
    pa.field("x", pa.float16()),
    pa.field("y", pa.float16()),
    pa.field("width", pa.float16()),
    pa.field("height", pa.float16()),
    # from eyepop.data.data_types import UserReview
    pa.field("user_review", pa.dictionary(pa.int8(), pa.string())),
]

OBJECT_STRUCT = pa.struct(_object_fields)
OBJECT_SCHEMA = pa.schema(_object_fields)

_class_fields = [
    # Python type is string, there should be only as small number of
    # classLabel values per prediction, hence the dictionary encoding.
    pa.field("classLabel", pa.dictionary(pa.int32(), pa.string())),
    pa.field("confidence", pa.float16()),
    # from eyepop.data.data_types import UserReview
    pa.field("user_review", pa.dictionary(pa.int8(), pa.string())),
]

CLASS_STRUCT = pa.struct(_class_fields)
CLASS_SCHEMA = pa.schema(_class_fields)

_annotation_fields = [
    # from eyepop.data.data_types import AnnotationType
    pa.field("type", pa.dictionary(pa.int8(), pa.string())),
    # from eyepop.data.data_types import AutoAnnotate
    pa.field("source", pa.dictionary(pa.int32(), pa.string())),
    # from eyepop.data.data_types import UserReview
    pa.field("user_review", pa.dictionary(pa.int8(), pa.string())),
    pa.field("objects", pa.list_(OBJECT_STRUCT)),
    pa.field("classes", pa.list_(CLASS_STRUCT))
]

ANNOTATION_STRUCT = pa.struct(_annotation_fields)

ANNOTATION_SCHEMA = pa.schema(_annotation_fields)

_asset_fields = [
    pa.field("uuid", pa.string()),
    pa.field("external_id", pa.string()),
    pa.field("created_at", pa.timestamp("ms")),
    pa.field("updated_at", pa.timestamp("ms")),
    pa.field("asset_url", pa.string()),
    pa.field("original_image_width", pa.uint16()),
    pa.field("original_image_height", pa.uint16()),
    # Python type is string, there should be only as small number of
    # partition values per asset list, hence the dictionary encoding.
    pa.field("partition", pa.dictionary(pa.int32(), pa.string())),
    pa.field("review_priority", pa.float16()),
    pa.field("model_relevance", pa.float16()),
    pa.field("annotations", pa.list_(ANNOTATION_STRUCT)),
]

ASSET_STRUCT = pa.struct(_asset_fields)

ASSET_SCHEMA = pa.schema(_asset_fields)



import pyarrow as pa

from eyepop.data.arrow.pandas.assets import dataframe_from_table, table_from_dataframe
from eyepop.data.arrow.schema import ASSET_SCHEMA as ASSET_SCHEMA_LATEST
from eyepop.data.arrow.schema_0_0 import ASSET_SCHEMA as ASSET_SCHEMA_0_0
from eyepop.data.arrow.schema_1_0 import ASSET_SCHEMA as ASSET_SCHEMA_1_0
from .arrow_test_helpers import create_test_table


def test_arrow_to_pandas():
    _test_arrow_to_pandas(ASSET_SCHEMA_LATEST)

def test_arrow_to_pandas_0_0():
    _test_arrow_to_pandas(ASSET_SCHEMA_0_0)

def test_arrow_to_pandas_1_0():
    _test_arrow_to_pandas(ASSET_SCHEMA_1_0)

def _test_arrow_to_pandas(schema: pa.Schema):
    table_in = create_test_table(schema=schema)
    dataframe = dataframe_from_table(table_in, schema=schema)
    table_out = table_from_dataframe(dataframe, schema=schema)
    assert table_in.schema == table_out.schema
    assert table_in.to_pylist() == table_out.to_pylist()




import asyncio
import io
import logging
import threading
import types
import typing
from asyncio import StreamReader

log = logging.getLogger(__name__)


class SyncEndpoint:
    def __init__(self, endpoint: "Endpoint"):
        self._on_ready = None
        self.endpoint = endpoint
        self.event_loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._run_event_loop, args=(self.event_loop,), daemon=True)
        self.thread.start()

    def __del__(self):
        self.event_loop.close()

    def __enter__(self) -> "SyncEndpoint":
        self.connect()
        return self

    def __exit__(
            self,
            exc_type: typing.Optional[typing.Type[BaseException]],
            exc_val: typing.Optional[BaseException],
            exc_tb: typing.Optional[types.TracebackType],
    ) -> None:
        self.disconnect()

    def connect(self):
        run_coro_thread_save(self.event_loop, self.endpoint.connect())

    def disconnect(self, timeout: float | None = None):
        run_coro_thread_save(self.event_loop, self.endpoint.disconnect(timeout))

    def session(self) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.session())

    def _run_event_loop(self, event_loop):
        log.debug("_run_event_loop start")
        asyncio.set_event_loop(event_loop)
        event_loop.run_forever()
        log.debug("_run_event_loop done")

    def _async_reader_to_sync_binary_io(self, async_stream_reader):
        queue = run_coro_thread_save(
            self.event_loop, _create_queue()
        )
        submit_coro_thread_save(
            self.event_loop, _drain_stream_reader_into_queue(async_stream_reader, queue)
        )
        sync_io = _async_queue_to_stream(self.event_loop, queue)
        return sync_io


def run_coro_thread_save(event_loop, coro):
    try:
        result = asyncio.run_coroutine_threadsafe(coro, event_loop).result()
        coro = None
        return result
    finally:
        if coro is not None:
            coro.close()


def submit_coro_thread_save(event_loop, coro):
    try:
        future = asyncio.run_coroutine_threadsafe(coro, event_loop)
        coro = None
        future.add_done_callback(lambda f: f.result())
    finally:
        if coro is not None:
            coro.close()


async def _create_queue() -> asyncio.Queue:
    return asyncio.Queue(maxsize=128)


async def _drain_stream_reader_into_queue(stream_reader: StreamReader, queue: asyncio.Queue):
    try:
        n = 0
        while True:
            buffer = await stream_reader.read(65536)
            if not buffer:
                break
            n += len(buffer)
            await queue.put(buffer)
    except Exception as e:
        await queue.put(e)
    finally:
        await queue.put(None)


# https://discuss.python.org/t/asynchronous-generator-to-io-bufferedreader/20503
# iterate over the request.stream(),
# putting the blocks into a queue
# have a separate thread that consumes the queue yielding bytes
# convert the generator into a buffered reader as per below>
# pass the buffered reader into tarfile.open(fileobj=...)

def _async_queue_to_stream(event_loop, queue: asyncio.Queue):
    class GeneratorStream(io.RawIOBase):
        def __init__(self):
            self.leftover = None
            self.eof = False

        def readable(self):
            return True

        def readinto(self, b):
            _l = len(b)  # : We're supposed to return at most this much
            chunk = self.leftover
            if not chunk and not self.eof:
                chunk = asyncio.run_coroutine_threadsafe(queue.get(), event_loop).result()
            if not chunk:
                self.eof = True
                return 0
            output, self.leftover = chunk[:_l], chunk[_l:]
            b[:len(output)] = output
            return len(output)
    return io.BufferedReader(GeneratorStream())



from pyarrow._compute import CastOptions
import pyarrow as pa

from . import schema_1_5 as schema_latest

""" The latest official Arrow schema for the EyePop Dataset API.

The latest officially supported schema is: 1.5

These are references to the types and schemas that are currently 
supported. For backward compatibility, we keep schemas versioned 
and immutable once officially supported.

Using schemas from newer-than-latest version is unstable. 
"""

MIME_TYPE_APACHE_ARROW_FILE_VERSIONED = schema_latest.MIME_TYPE_APACHE_ARROW_FILE_VERSIONED

OBJECT_STRUCT = schema_latest.OBJECT_STRUCT
OBJECT_SCHEMA = schema_latest.OBJECT_SCHEMA

CLASS_STRUCT = schema_latest.CLASS_STRUCT
CLASS_SCHEMA = schema_latest.CLASS_SCHEMA

KEY_POINT_STRUCT = schema_latest.KEY_POINT_STRUCT
KEY_POINT_SCHEMA = schema_latest.KEY_POINT_SCHEMA

KEY_POINTS_STRUCT = schema_latest.KEY_POINTS_STRUCT
KEY_POINTS_SCHEMA = schema_latest.KEY_POINTS_SCHEMA

TEXT_STRUCT = schema_latest.TEXT_STRUCT
TEXT_SCHEMA = schema_latest.TEXT_SCHEMA

EMBEDDING_STRUCT = schema_latest.EMBEDDING_STRUCT
EMBEDDING_SCHEMA = schema_latest.EMBEDDING_SCHEMA

ANNOTATION_STRUCT = schema_latest.ANNOTATION_STRUCT
ANNOTATION_SCHEMA = schema_latest.ANNOTATION_SCHEMA

ASSET_STRUCT = schema_latest.ASSET_STRUCT
ASSET_SCHEMA = schema_latest.ASSET_SCHEMA



import json
from asyncio import Queue
from typing import Callable, BinaryIO
from urllib.parse import quote_plus

import aiohttp

from eyepop.client_session import ClientSession
from eyepop.data.data_types import Asset, AssetImport
from eyepop.jobs import Job, JobStateCallback


class DataJob(Job):
    """
    Abstract Job submitted to an EyePop.ai DataEndpoint.
    """

    def __init__(self, session: ClientSession, on_ready: Callable[["DataJob"], None] | None,
                 callback: JobStateCallback | None = None):
        super().__init__(session, on_ready, callback)

    async def result(self) -> Asset:
        return await self.pop_result()

    async def _do_execute_job(self, queue: Queue, session: ClientSession):
        raise NotImplementedError("can't execute abstract jobs")


class _UploadStreamJob(DataJob):
    def __init__(
            self,
            stream: BinaryIO,
            mime_type: str,
            dataset_uuid: str,
            dataset_version: int | None,
            external_id: str | None,
            sync_transform: bool | None,
            no_transform: bool | None,
            session: ClientSession,
            on_ready: Callable[[DataJob], None] | None = None,
            callback: JobStateCallback | None = None
    ):
        super().__init__(session, on_ready, callback)
        self.stream = stream
        self.mime_type = mime_type
        self.dataset_uuid = dataset_uuid
        self.dataset_version = dataset_version
        self.external_id = external_id
        self.sync_transform = sync_transform
        self.no_transform = no_transform

    async def _do_execute_job(self, queue: Queue, session: ClientSession):
        dataset_version_query = f"&dataset_version={self.dataset_version}" if self.dataset_version else ""
        external_id_query = f"&external_id={quote_plus(self.external_id)}" if self.external_id else ""

        post_path = f"/assets?dataset_uuid={self.dataset_uuid}{dataset_version_query}{external_id_query}"

        if self.sync_transform is not None:
            post_path = f"{post_path}&sync_transform={'true' if self.sync_transform else 'false'}"
        if self.no_transform is not None:
            post_path = f"{post_path}&no_transform={'true' if self.no_transform else 'false'}"

        async with await session.request_with_retry(
                method="POST",
                url=post_path,
                data=self.stream,
                content_type=self.mime_type,
                timeout=aiohttp.ClientTimeout(total=None, sock_read=60)
        ) as resp:
            result = Asset.model_validate(await resp.json())
            await queue.put(result)


class _ImportFromJob(DataJob):
    def __init__(
            self,
            asset_import: AssetImport,
            dataset_uuid: str,
            dataset_version: int | None,
            external_id: str | None,
            partition: str | None,
            sync_transform: bool | None,
            no_transform: bool | None,
            session: ClientSession,
            on_ready: Callable[[DataJob], None] | None = None,
            callback: JobStateCallback | None = None
    ):
        super().__init__(session, on_ready, callback)
        self.asset_import = asset_import
        self.dataset_uuid = dataset_uuid
        self.dataset_version = dataset_version
        self.external_id = external_id
        self.partition = partition
        self.sync_transform = sync_transform
        self.no_transform = no_transform

    async def _do_execute_job(self, queue: Queue, session: ClientSession):
        dataset_version_query = f"&dataset_version={self.dataset_version}" if self.dataset_version else ""
        external_id_query = f"&external_id={quote_plus(self.external_id)}" if self.external_id else ""
        partition_query = f"&partition={quote_plus(self.partition)}" if self.partition else ""

        post_path = (f"/assets/imports?dataset_uuid={self.dataset_uuid}"
                     f"{dataset_version_query}{external_id_query}{partition_query}")

        if self.sync_transform is not None:
            post_path = f"{post_path}&sync_transform={'true' if self.sync_transform else 'false'}"
        if self.no_transform is not None:
            post_path = f"{post_path}&no_transform={'true' if self.no_transform else 'false'}"

        async with await session.request_with_retry(
                "POST",
                post_path,
                data=self.asset_import.model_dump_json(exclude_unset=True),
                content_type="application/json",
                timeout=aiohttp.ClientTimeout(total=None, sock_read=60)) as resp:
            result = Asset.model_validate(await resp.json())
            await queue.put(result)



import enum
from datetime import datetime
from typing import Any, Dict, List, Callable, Awaitable

from pydantic import BaseModel, ConfigDict, Field


class AssetStatus(enum.StrEnum):
    rejected = enum.auto()
    upload_in_progress = enum.auto()
    transform_in_progress = enum.auto()
    enrich_in_progress = enum.auto()
    accepted = enum.auto()


class TranscodeMode(enum.StrEnum):
    original = enum.auto()
    image_original_size = enum.auto()
    image_fit_1024 = enum.auto()
    image_fit_640 = enum.auto()
    image_fit_224 = enum.auto()
    image_cover_1024 = enum.auto()
    image_cover_640 = enum.auto()
    image_cover_224 = enum.auto()


AutoAnnotate = str


class ModelType(enum.StrEnum):
    epdet_b1 = enum.auto()
    epdet_b1_1 = enum.auto()
    yolov7 = enum.auto()
    yolov7_tiny = enum.auto()
    yolov7_e6e = enum.auto()
    yolov7_w6 = enum.auto()
    yolov7_x = enum.auto()
    coolr_demo = enum.auto()
    imported = enum.auto()
    eyepop_sst = enum.auto()


class ModelStatus(enum.StrEnum):
    error = enum.auto()
    draft = enum.auto()
    requested = enum.auto()
    in_progress = enum.auto()
    available = enum.auto()
    published = enum.auto()


class AnnotationType(enum.StrEnum):
    ground_truth = enum.auto()
    prediction = enum.auto()
    auto = enum.auto()


class UserReview(enum.StrEnum):
    approved = enum.auto()
    rejected = enum.auto()
    unknown = enum.auto()


class DatasetVersionAssetStats(BaseModel):
    total: int | None = None
    accepted: int | None = None
    annotated: int | None = None
    auto_annotated: int | None = None
    auto_annotated_approved: int | None = None
    ground_truth_annotated: int | None = None


class DatasetVersion(BaseModel):
    version: int
    created_at: datetime | None = None
    updated_at: datetime | None = None
    assets_modified_at: datetime | None = None
    last_analysed_at: datetime | None = None
    modifiable: bool = False
    hero_asset_uuid: str | None = None
    asset_stats: DatasetVersionAssetStats | None = None


DatasetVersionResponse = DatasetVersion

class AutoAnnotateTask(enum.StrEnum):
    object_detection = enum.auto()
    image_classification = enum.auto()


class AutoAnnotatePrompt(BaseModel):
    label: str
    prompt: str


class AutoAnnotateParams(BaseModel):
    # @deprecated("candidate_labels is deprecated, use prompts instead")
    candidate_labels: list[str] | None = None
    prompts: list[AutoAnnotatePrompt] | None = None
    confidence_threshold: float | None = None
    task: AutoAnnotateTask = AutoAnnotateTask.object_detection


class DatasetParent(BaseModel):
    dataset_uuid: str
    dataset_version: int


class Dataset(BaseModel):
    uuid: str
    name: str
    description: str = ""
    created_at: datetime | None = None
    updated_at: datetime | None = None
    tags: List[str] = []
    account_uuid: str
    auto_annotates: List[AutoAnnotate]
    auto_annotate_params: AutoAnnotateParams | None = None
    versions: List[DatasetVersion] = []
    modifiable_version: int | None = None
    parent: DatasetParent | None = None
    searchable: bool | None = None


DatasetResponse = Dataset


class DatasetCreate(BaseModel):
    name: str
    description: str = ""
    tags: list[str] = []
    auto_annotates: list[AutoAnnotate] = []
    auto_annotate_params: AutoAnnotateParams | None = None
    searchable: bool | None = None
    model_config = ConfigDict(extra='forbid')


class DatasetUpdate(DatasetCreate):
    name: str | None = None
    description: str | None = None
    tags: List[str] | None = None
    auto_annotates: List[AutoAnnotate]  | None = None
    auto_annotate_params: AutoAnnotateParams | None = None
    searchable: bool | None = None
    model_config = ConfigDict(extra='forbid')


class Point2d(BaseModel):
    x: float
    y: float


class Point3d(Point2d):
    z: float | None = None


class Box(BaseModel):
    topLeft: Point2d
    bottomRight: Point2d


class PredictedClass(BaseModel):
    id: int | None = None
    confidence: float | None = None
    classLabel: str
    category: str | None = None

class PredictedEmbedding(BaseModel):
    x: float | None = None
    y: float | None = None
    embedding: List[float]
    category: str | None = None

class PredictedText(BaseModel):
    id: int | None = None
    confidence: float | None = None
    text: str
    category: str | None = None


class Contour(BaseModel):
    points: List[Point2d]
    cutouts: List[List[Point2d]]


class Mask(BaseModel):
    bitmap: str
    width: int
    height: int
    stride: int


class PredictedMesh(BaseModel):
    id: int | None = None
    category: str | None = None
    confidence: float = None
    points: List[Point3d]


class PredictedKeyPoint(Point3d):
    id: int | None = None
    confidence: float | None = None
    classLabel: str | None = None
    category: str | None = None
    visible: bool | None = None


class PredictedKeyPoints(BaseModel):
    category: str | None = None
    type: str | None = None
    points: List[PredictedKeyPoint]


class PredictedObject(PredictedClass):
    trackId: int | None = None
    x: float
    y: float
    width: float
    height: float
    orientation: float | None = None
    outline: List[Point2d] | None = None
    contours: List[Contour] | None = None
    mask: Mask | None = None
    objects: List["PredictedObject"] | None = None
    classes: List[PredictedClass] | None = None
    texts: List[PredictedText] | None = None
    meshs: List[PredictedMesh] | None = None
    keyPoints: List[PredictedKeyPoints] | None = None


class Prediction(BaseModel):
    """Represents a prediction for an asset or a chunk of an asset."""
    source_width: float = Field(
        description="The width of the source coordinate system for all prediction coordinates."
    )
    source_height: float = Field(
        description="The height of the source coordinate system for all prediction coordinates."
    )
    timestamp: int | None = Field(
        default=None,
        description="Temporal offset of prediction from start of the media (video or audio) in nano seconds."
    )
    duration: int | None = Field(
        default=None,
        description="Temporal length of the chunk that was the source for this prediction in nano seconds."
    )
    offset: int | None = Field(
        default=None,
        description="A media specific offset. For video frames, this is the frame number of prediction. "
                    "For audio samples, this is the offset of the first sample for this prediction."
    )
    offset_duration: int | None = Field(
        default=None,
        description="Offset length of the chunk used for this prediction. It has the same format as offset."
    )

    objects: List[PredictedObject] | None = None
    classes: List[PredictedClass] | None = None
    texts: List[PredictedText] | None = None
    meshs: List[PredictedMesh] | None = None
    keyPoints: List[PredictedKeyPoints] | None = None
    embeddings: List[PredictedEmbedding] | None = None


class AssetAnnotation(BaseModel):
    type: AnnotationType
    user_review: UserReview = UserReview.unknown
    approved_threshold: float | None = None
    auto_annotate: AutoAnnotate | None = None
    auto_annotate_params: AutoAnnotateParams | None = None
    source: str | None = None
    annotation: Prediction | None = None
    uncertainty_score: float | None = None
    source_model_uuid: str | None = None


AssetAnnotationResponse = AssetAnnotation


class Asset(BaseModel):
    uuid: str
    created_at: datetime | None = None
    updated_at: datetime | None = None
    mime_type: str | None = None
    file_size_bytes: int | None = None
    is_transformed: bool | None = None
    status: AssetStatus | None = None
    status_message: str | None = None
    external_id: str | None = None
    partition: str | None = None
    review_priority: float | None = None
    model_relevance: float | None = None
    annotations: List[AssetAnnotation] = []


    ###
    # Denormalized attributes, for convenient serialization/storage detached of a dataset context
    # ###
    dataset_uuid: str | None = None
    account_uuid: str | None = None

    model_config = ConfigDict(
        protected_namespaces=('pydantic_do_not_prevent_model_prefix_',)
    )


AssetResponse = Asset


class AssetImport(BaseModel):
    url: str
    ground_truth: Prediction | None = None


class ModelMetrics(BaseModel):
    cpr: list[tuple[float, float, float]] | None = None


class ModelExportFormat(enum.StrEnum):
    TensorFlowLite = "TensorFlowLite"
    TensorFlowGraphDef = "TensorFlowGraphDef"
    TorchScript = "TorchScript"
    TorchScriptCpu = "TorchScriptCpu"
    TorchScriptCuda = "TorchScriptCuda"
    ONNX = "ONNX"
    PyTorch = "PyTorch"
    ModelLess = "ModelLess"


class ModelExportStatus(enum.StrEnum):
    in_progress = enum.auto()
    finished = enum.auto()
    error = enum.auto()


class ExportedBy(enum.StrEnum):
    eyepop = enum.auto()
    qc_ai_hub = enum.auto()


class ModelExport(BaseModel):
    format: ModelExportFormat
    exported_by: ExportedBy
    export_params: dict[str, str] | None = None
    status: ModelExportStatus


ModelExportResponse = ModelExport


class ModelTask(enum.StrEnum):
    object_detection = enum.auto()
    image_classification = enum.auto()
    keypoint_detection = enum.auto()


class Model(BaseModel):
    uuid: str
    created_at: datetime | None = None
    updated_at: datetime | None = None
    account_uuid: str
    dataset_uuid: str | None = None
    dataset_version: int | None = None
    name: str
    description: str = ""
    type: ModelType | None = None
    is_public: bool = False
    external_id: str | None = None
    pretrained_model_uuid: str | None = None
    extra_params: dict | None = None
    task: ModelTask | None = None
    classes: list[str] | None = None
    status: ModelStatus = ModelStatus.draft
    status_message: str | None = None
    metrics: ModelMetrics | None = None
    exports: list[ModelExport] | None = None


ModelResponse = Model


class ModelCreate(BaseModel):
    name: str
    description: str
    external_id: str | None = None
    pretrained_model_uuid: str | None = None
    extra_params: dict | None = None
    task: ModelTask | None = None
    classes: list[str] | None = None
    type: ModelType | None = None
    status: ModelStatus | None = None


class ModelUpdate(ModelCreate):
    name: str | None = None
    description: str | None = None
    is_public: bool | None = None
    external_id: str | None = None
    status: ModelStatus | None = None
    task: ModelTask | None = None
    classes: list[str] | None = None


class ModelTrainingStage(enum.StrEnum):
    waiting = enum.auto()
    scheduling = enum.auto()
    preparing = enum.auto()
    training = enum.auto()
    exporting = enum.auto()


class ModelSample(BaseModel):
    asset_uuid: str
    prediction: Prediction


class ModelTrainingProgress(BaseModel):
    stage: ModelTrainingStage
    queue_length: int | None = None
    started_at: datetime | None = None
    finished_at: datetime | None = None
    status_code: int | None = None
    status_message: str | None = None
    metrics: ModelMetrics | None = None
    samples: list[ModelSample] | None = None
    remaining_seconds_min: float | None = None
    remaining_seconds_max: float | None = None


class ChangeType(enum.StrEnum):
    dataset_added = enum.auto()
    dataset_removed = enum.auto()
    dataset_modified = enum.auto()
    dataset_version_modified = enum.auto()
    asset_added = enum.auto()
    asset_removed = enum.auto()
    asset_status_modified = enum.auto()
    asset_annotation_modified = enum.auto()
    model_added = enum.auto()
    model_removed = enum.auto()
    model_modified = enum.auto()
    model_status_modified = enum.auto()
    model_progress = enum.auto()
    events_lost = enum.auto()
    
    workflow_started = enum.auto()
    workflow_succeeded = enum.auto()
    workflow_failed = enum.auto()

    workflow_task_started = enum.auto()
    workflow_task_succeeded = enum.auto()
    workflow_task_failed = enum.auto()


class ChangeEvent(BaseModel):
    change_type: ChangeType
    account_uuid: str
    dataset_uuid: str | None
    dataset_version: int | None
    asset_uuid: str | None
    mdl_uuid: str | None
    workflow_id: str | None
    message: str | None
    workflow_task_name: str | None



EventHandler = Callable[[ChangeEvent], Awaitable[None]]


class Tag(BaseModel):
    name: str
    model_uuid: str
    model_config = ConfigDict(
        protected_namespaces=('pydantic_do_not_prevent_model_prefix_',)
    )


TagResponse = Tag


class ModelAlias(BaseModel):
    name: str
    description: str | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    account_uuid: str
    is_public: bool = False
    tags: list[Tag] = []


ModelAliasResponse = ModelAlias


class ModelAliasCreate(BaseModel):
    name: str
    description: str | None = None
    is_public: bool = False


class ModelAliasUpdate(BaseModel):
    description: str | None = None
    is_public: bool | None = None


class AliasResolution(BaseModel):
    alias: str
    model_uuid: str | None = None
    model_config = ConfigDict(
        protected_namespaces=('pydantic_do_not_prevent_model_prefix_',)
    )


ExportedAliasResponse = AliasResolution

MIME_TYPE_APACHE_ARROW_FILE = "application/vnd.apache.arrow.file"


class AssetUrlType(enum.StrEnum):
    gcs = enum.auto()
    s3 = enum.auto()
    https_signed = enum.auto()


class AssetInclusionMode(enum.StrEnum):
    all_assets = enum.auto()
    annotated_only = enum.auto()
    manual_annotated_only = enum.auto()
    auto_annotated_only = enum.auto()


class AnnotationInclusionMode(enum.StrEnum):
    all = enum.auto()
    user_reviewed = enum.auto()
    user_approved = enum.auto()
    ground_truth = enum.auto()


class ModelTrainingStatusCode(enum.Enum):
    ok = 200
    bad_input_error = 400
    internal_error = 500


class ModelTrainingEvent(BaseModel):
    """ Model training event.

        This structure is used for the model training process to communicate the status events.
        The model training workflow SHOULD send at least two updates:

        1. At the beginning of each stage, all attributes except `stage` should be none. The actual
           start time of the stage will be recorded automatically based on the clock time of the receiver.

        2. At the end of each stage, with the same `stage` value and a `status_code`. Any `status_code`
           other than 200 will be interpreted as failure. In this case, the sender should send a huma
           readable, `status_message` for developers to debug the root cause.

        The model training workflow MAY send regular intermediary updates if it can estimate the remaining
        time it will take to complete the current stage. It does so by passing `elapsed_units`,
        `remaining_units_min` as lower bound and optionally `remaining_units_max` as upper bound. The units
        used can be freely chosen by the sender, for example it can be `elapsed seconds`, `used GPU cycles`,
        `epochs`, `batches` or any other unit that is appropriate to express the remaining "work" left,
        relative to the "work" already spend.

        The model training workflow MAY include:
        * `samples` a list of tuples (`asset_uuid`, `prediction`) for a small number of representative
           asset. The prediction are meant to illustrate the current performance of the model in training
           on a set of representative assets.
        * `cpr` A graph described as a list of tuples (`confidence`, `precision`, `recall`) to illustrate
          the current model performance against the validation partition.
        * OR `cpr_transposed` a convenience attribute to send the same data just transposed.
        * `job_identifier` to set an external job identifier if not set already, this will be ignored
          if the external job identifier is already set.
        * `work_storage_url` to set the work storage url if not set already, this will be ignored
          if the work storage url is already set.

    """
    stage: ModelTrainingStage
    job_identifier: str | None = None
    work_storage_url: str | None = None
    elapsed_units: float | None = None
    remaining_units_min: float | None = None
    remaining_units_max: float | None = None
    status_code: ModelTrainingStatusCode | None = None
    status_message: str | None = None
    samples: list[tuple[str, Prediction]] | None = None
    cpr: list[tuple[float, float, float]] | None = None
    cpr_transposed: tuple[list[float], list[float], list[float]] | None = None


class ModelTrainingAuditRecord(BaseModel):
    """ Internal audit record for model trainings.
    """
    model_uuid: str
    model: ModelResponse
    work_storage_url: str | None = None
    model_config = ConfigDict(
        protected_namespaces=('pydantic_do_not_prevent_my_prefix_',),
    )


class ExportedUrlResponse(BaseModel):
    model_uuid: str
    model_format: ModelExportFormat
    exported_url: str
    model_config = ConfigDict(
        protected_namespaces=('pydantic_do_not_prevent_my_prefix_',)
    )


class ArtifactType(enum.StrEnum):
    eyepop_bundle = enum.auto()
    weights_file = enum.auto()


class QcAiHubExportParams(BaseModel):
    device_name: str


class ArgoWorkflowPhase(enum.StrEnum):
    UNKNOWN = "Unknown"
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    ERROR = "Error"


class CreateWorkflowBody(BaseModel):
    parameters: dict | None = None

class CreateWorkflowParameters(BaseModel):
    dataset_uuid: str | None = None
    dataset_version: int | None = None
    model_uuid: str | None = None
    config: Dict[str, Any] | None = None
    root_base_url: str | None = None

class CreateWorkflowResponse(BaseModel):
    workflow_id: str


class ListWorkflowItemMetadataLabels(BaseModel):
    account_uuid: str
    dataset_uuid: str | None = None
    model_uuid: str | None = None
    phase: ArgoWorkflowPhase


class ListWorkflowItemMetadata(BaseModel):
    workflow_id: str
    created_at: datetime
    labels: ListWorkflowItemMetadataLabels


class ListWorkflowItem(BaseModel):
    metadata: ListWorkflowItemMetadata



import json
import unittest
from aioresponses import aioresponses

from eyepop import EyePopSdk
from .base_endpoint_test import BaseEndpointTest


class TestEndpointConnect(BaseEndpointTest):

    @aioresponses()
    def test_no_tracing(self, mock: aioresponses):

        self.setup_base_mock(mock)

        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        endpoint = EyePopSdk.dataEndpoint(
            eyepop_url=self.test_eyepop_url,
            secret_key=self.test_eyepop_secret_key,
            account_id=self.test_eyepop_account_id,
            request_tracer_max_buffer=0
        )
        try:
            endpoint.connect()
        finally:
            endpoint.disconnect()

        self.assertBaseMock(mock)


if __name__ == '__main__':
    unittest.main()



from eyepop.data.data_types import AssetAnnotationResponse, Prediction, PredictedObject, PredictedClass, PredictedEmbedding

# Confidence and coordinates are represented as float16 in the arrow format but Python lacks support for 2-byte floats.
# To avoid "changing" 4-bytes floats when converted back and forth, we will always round to these precisions.

CONFIDENCE_N_DIGITS = 3
COORDINATE_N_DIGITS = 3
EMBEDDING_N_DIGITS = 1

def normalize_eyepop_annotations(
        annotations: list[AssetAnnotationResponse]
):
    for annotation in annotations:
        if annotation.annotation is not None:
            normalize_eyepop_prediction(annotation.annotation)


def normalize_eyepop_prediction(
        prediction: Prediction
):
    if prediction.objects:
        normalize_predicted_objects(
            prediction.objects, prediction.source_width, prediction.source_height)
    if prediction.classes:
        normalize_predicted_classes(prediction.classes)
    if prediction.embeddings:
        normalize_predicted_embeddings(prediction.embeddings)
    prediction.source_width = 1.0
    prediction.source_height = 1.0


def normalize_predicted_objects(
        predicted_objects: list[PredictedObject],
        original_source_width: float,
        original_source_height: float
):
    for o in predicted_objects:
        if o.objects:
            normalize_predicted_objects(o.objects, original_source_width, original_source_height)
        if o.classes:
            normalize_predicted_classes(o.classes)
        o.x = round(o.x / original_source_width, COORDINATE_N_DIGITS)
        o.y = round(o.y / original_source_height, COORDINATE_N_DIGITS)
        o.width = round(o.width / original_source_width, COORDINATE_N_DIGITS)
        o.height = round(o.height / original_source_height, COORDINATE_N_DIGITS)
        if o.confidence is not None:
            o.confidence = round(o.confidence, CONFIDENCE_N_DIGITS)


def normalize_predicted_classes(
        predicted_classes: list[PredictedClass]
):
    for c in predicted_classes:
        if c.confidence is not None:
            c.confidence = round(c.confidence, CONFIDENCE_N_DIGITS)


def normalize_predicted_embeddings(
        predicted_embeddings: list[PredictedEmbedding]
):
    for e in predicted_embeddings:
        if e.embedding is not None:
            for i in range(len(e.embedding)):
                e.embedding[i] = round(e.embedding[i], EMBEDDING_N_DIGITS)



import asyncio
import json
import logging
from io import StringIO
from typing import BinaryIO
from uuid import uuid4

import aiohttp

from eyepop.data.data_endpoint import DataEndpoint
from eyepop.data.data_types import ModelCreate, ModelExportFormat, ModelUpdate, ModelStatus, ModelAliasCreate
from eyepop.eyepopsdk import EyePopSdk
from eyepop.worker.worker_endpoint import WorkerEndpoint
from eyepop.worker.worker_types import Pop, InferenceComponent, InferenceType
from importlib import resources
from examples.experimental import sample_assets

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.getLogger('eyepop.requests').setLevel(level=logging.DEBUG)

log = logging.getLogger(__name__)

LABELS_URL = 'https://s3.amazonaws.com/models.eyepop.ai/releases/yolov7/1.0.1/models/YOLOv7/COCO/Latest/COCO_Latest_labels.txt'
WEIGHTS_URL = 'https://s3.amazonaws.com/models.eyepop.ai/releases/yolov7/1.0.1/models/YOLOv7/COCO/Latest/TensorFlowLite/float32/yolov7_YOLOv7_COCO_Latest_TensorFlowLite_float32.tflite'
MODEL_JSON_URL = 'https://s3.amazonaws.com/models.eyepop.ai/releases/yolov7/1.0.1/models/YOLOv7/COCO/Latest/TensorFlowLite/float32/model.json'

ALIAS = f"{uuid4().hex}.example.eyepop.ai"
TAG = "latest"

async def import_artifacts(endpoint: DataEndpoint, model_uuid: str):
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        log.info("importing from: %s", LABELS_URL)
        async with session.get(LABELS_URL) as resp:
            await endpoint.upload_model_artifact(
                model_uuid,
                ModelExportFormat.TensorFlowLite,
                'labels.txt',
                resp.content
            )
        log.info("importing from: %s", WEIGHTS_URL)
        async with session.get(WEIGHTS_URL) as resp:
            await endpoint.upload_model_artifact(
                model_uuid,
                ModelExportFormat.TensorFlowLite,
                'yolov7_YOLOv7_COCO_Latest_TensorFlowLite_float32.tflite',
                resp.content
            )
        log.info("importing from: %s", MODEL_JSON_URL)
        async with session.get(MODEL_JSON_URL) as resp:
            model_json = json.loads(await resp.content.read())
            # patch the model.json because we upload the labels.txt in a different location
            model_json['assets'][1]['url'] = "labels.txt"
            await endpoint.upload_model_artifact(
                model_uuid,
                ModelExportFormat.TensorFlowLite,
                'yolov7_YOLOv7_COCO_Latest_TensorFlowLite_float32.tflite',
                json.dumps(model_json),
                mime_type='application/json'
            )

async def finish_model(endpoint: DataEndpoint, model_uuid: str):
    await endpoint.update_model(model_uuid, ModelUpdate(status=ModelStatus.available))
    log.info("made available: %s", model_uuid)

async def set_model_alias(endpoint: DataEndpoint, model_uuid: str, alias: str, tag: str):
    await endpoint.publish_model(model_uuid)
    await endpoint.create_model_alias(ModelAliasCreate(name=alias))
    log.info("created alias: %s", alias)
    await endpoint.set_model_alias_tag(alias, tag, model_uuid)
    log.info("set alais %s:%s to %s", alias, tag, model_uuid)

async def use_model(worker: WorkerEndpoint, model_uuid: str):
    await worker.set_pop(Pop(
        components=[InferenceComponent(
            inferenceTypes=[InferenceType.OBJECT_DETECTION],
            modelUuid=model_uuid
        )]
    ))
    example_file = resources.files(sample_assets) / "exampple.jpg"
    with example_file.open("rb") as f:
        job = await worker.upload_stream(f, "image/jpeg")
        result = job.predict()
        log.info("inference result: %s", result)

async def main():
    async with EyePopSdk.dataEndpoint(is_async=True, disable_ws=False) as endpoint:
        model = await endpoint.create_model(ModelCreate(
            name="test model",
            description=""
        ))
        try:
            log.info("created model: %s", model.uuid)
            await import_artifacts(endpoint, model.uuid)
            await finish_model(endpoint, model.uuid)
            try:
                await set_model_alias(endpoint, model.uuid, ALIAS, TAG)
                log.info("model ready to use: %s - starting a pop now", model.uuid)
                async with EyePopSdk.workerEndpoint(pop_id='transient', is_async=True) as worker:
                    await use_model(worker, model.uuid)
            finally:
                await endpoint.delete_model_alias_tag(ALIAS, TAG)
                await endpoint.delete_model_alias(ALIAS)
        finally:
            await endpoint.delete_model(model.uuid)

asyncio.run(main())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: events

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Event(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsEvent(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Event()
        x.Init(buf, n + offset)
        return x

    # Event
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Event
    def Method(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Event
    def EventTimeEpochMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Event
    def XRequestId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def Result(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Event
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # Event
    def HostIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # Event
    def PathIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # Event
    def WaitMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Event
    def ProcessMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Event
    def BodyBytesSent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Event
    def BodyBytesReceived(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def EventStart(builder): builder.StartObject(11)
def EventAddMethod(builder, method): builder.PrependInt8Slot(0, method, 0)
def EventAddEventTimeEpochMs(builder, eventTimeEpochMs): builder.PrependUint64Slot(1, eventTimeEpochMs, 0)
def EventAddXRequestId(builder, xRequestId): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(xRequestId), 0)
def EventAddResult(builder, result): builder.PrependInt8Slot(3, result, 0)
def EventAddStatus(builder, status): builder.PrependInt16Slot(4, status, 0)
def EventAddHostIndex(builder, hostIndex): builder.PrependInt16Slot(5, hostIndex, 0)
def EventAddPathIndex(builder, pathIndex): builder.PrependInt16Slot(6, pathIndex, 0)
def EventAddWaitMs(builder, waitMs): builder.PrependUint32Slot(7, waitMs, 0)
def EventAddProcessMs(builder, processMs): builder.PrependUint32Slot(8, processMs, 0)
def EventAddBodyBytesSent(builder, bodyBytesSent): builder.PrependUint32Slot(9, bodyBytesSent, 0)
def EventAddBodyBytesReceived(builder, bodyBytesReceived): builder.PrependUint32Slot(10, bodyBytesReceived, 0)
def EventEnd(builder): return builder.EndObject()



import json
import time
from importlib import resources

import pytest
from aioresponses import aioresponses, CallbackResult

from eyepop import EyePopSdk
from eyepop.worker.worker_types import Pop
from tests.worker.base_endpoint_test import BaseEndpointTest
import tests

class TestEndpointUpload(BaseEndpointTest):
    test_source_id = 'test_source_id'
    test_file = resources.files(tests) / 'test.jpg'
    test_content_type = 'image/jpeg'


    @aioresponses()
    def test_sync_upload_ok(self, mock: aioresponses):
        self.setup_base_mock(mock)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pop(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': Pop(components=[]).model_dump()
}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pop)
        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id=self.test_eyepop_pop_id) as endpoint:
            self.assertBaseMock(mock)
            test_timestamp = time.time() * 1000 * 1000 * 1000

            upload_called = 0

            def upload(url, **kwargs) -> CallbackResult:
                nonlocal test_timestamp
                nonlocal upload_called
                upload_called += 1
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                elif kwargs['headers']['Content-Type'] != self.test_content_type:
                    return CallbackResult(status=40, reason='unsupported content type')
                else:
                    return CallbackResult(status=200,
                                          body=json.dumps(
                                              {'source_id': self.test_source_id, 'seconds': 0,
                                               'system_timestamp': test_timestamp}))

            mock.post(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/source?mode=queue&processing=sync',
                       callback=upload)

            job = endpoint.upload(self.test_file)
            result = job.predict()
            self.assertIsNotNone(result)
            self.assertEqual(result['source_id'], self.test_source_id)
            self.assertEqual(result['seconds'], 0)
            self.assertEqual(result['system_timestamp'], test_timestamp)
            self.assertIsNone(job.predict())

            self.assertEqual(upload_called, 1)


    @aioresponses()
    @pytest.mark.asyncio
    async def test_async_upload_ok(self, mock: aioresponses):
        self.setup_base_mock(mock)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pop(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': Pop(components=[]).model_dump()
}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pop)
        async with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                            pop_id=self.test_eyepop_pop_id, is_async=True) as endpoint:
            self.assertBaseMock(mock)
            test_timestamp = time.time() * 1000 * 1000 * 1000

            upload_called = 0

            def upload(url, **kwargs) -> CallbackResult:
                nonlocal test_timestamp
                nonlocal upload_called
                upload_called += 1
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                elif kwargs['headers']['Content-Type'] != self.test_content_type:
                    return CallbackResult(status=40, reason='unsupported content type')
                else:
                    return CallbackResult(status=200,
                                          body=json.dumps(
                                              {'source_id': self.test_source_id, 'seconds': 0,
                                               'system_timestamp': test_timestamp}))

            mock.post(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/source?mode=queue&processing=sync',
                       callback=upload)

            job = await endpoint.upload(self.test_file)
            result = await job.predict()
            self.assertIsNotNone(result)
            self.assertEqual(result['source_id'], self.test_source_id)
            self.assertEqual(result['seconds'], 0)
            self.assertEqual(result['system_timestamp'], test_timestamp)
            self.assertIsNone(await job.predict())

            self.assertEqual(upload_called, 1)




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: events

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Record(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRecord(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Record()
        x.Init(buf, n + offset)
        return x

    # Record
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Record
    def ClientType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Record
    def ClientVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Record
    def Hosts(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Record
    def HostsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Record
    def HostsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Record
    def Paths(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Record
    def PathsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Record
    def PathsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Record
    def Events(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from eyepop.events.Event import Event
            obj = Event()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Record
    def EventsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Record
    def EventsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def RecordStart(builder): builder.StartObject(5)
def RecordAddClientType(builder, clientType): builder.PrependInt8Slot(0, clientType, 0)
def RecordAddClientVersion(builder, clientVersion): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(clientVersion), 0)
def RecordAddHosts(builder, hosts): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(hosts), 0)
def RecordStartHostsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RecordAddPaths(builder, paths): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(paths), 0)
def RecordStartPathsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RecordAddEvents(builder, events): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(events), 0)
def RecordStartEventsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RecordEnd(builder): return builder.EndObject()



import logging
import time
from io import StringIO
from typing import Callable, BinaryIO, Any
from urllib.parse import urljoin

import aiohttp

from eyepop.endpoint import Endpoint
from eyepop.exceptions import PopNotStartedException, PopConfigurationException, PopNotReachableException
from eyepop.worker.worker_client_session import WorkerClientSession
from eyepop.worker.worker_jobs import WorkerJob, _UploadFileJob, _LoadFromJob, _UploadStreamJob, _LoadFromAssetUuidJob
from eyepop.worker.load_balancer import EndpointLoadBalancer
from eyepop.worker.worker_syncify import SyncWorkerJob
from eyepop.worker.worker_types import Pop, VideoMode, ComponentParams

log = logging.getLogger('eyepop')
log_requests = logging.getLogger('eyepop.requests')
log_metrics = logging.getLogger('eyepop.metrics')

MIN_CONFIG_RECONNECT_SECS = 10.0
MAX_RETRY_TIME_SECS = 30.0
FORCE_REFRESH_CONFIG_SECS = (61.0 * 61.0)


class WorkerEndpoint(Endpoint, WorkerClientSession):
    """
    Endpoint to an EyePop.ai worker.
    """

    def __init__(
            self,
            secret_key: str | None,
            access_token: str | None,
            eyepop_url: str,
            pop_id: str,
            auto_start: bool,
            stop_jobs: bool,
            job_queue_length: int,
            request_tracer_max_buffer: int,
            dataset_uuid: str | None = None,
    ):
        super().__init__(
            secret_key=secret_key,
            access_token=access_token,
            eyepop_url=eyepop_url,
            job_queue_length=job_queue_length,
            request_tracer_max_buffer=request_tracer_max_buffer
        )
        self.pop_id = pop_id
        self.auto_start = auto_start
        self.stop_jobs = stop_jobs
        self.dataset_uuid = dataset_uuid

        self.is_dev_mode = True

        self.worker_config = None
        self.last_fetch_config_success_time = None
        self.last_fetch_config_error = None
        self.last_fetch_config_error_time = None

        # new way
        self.pop = Pop(components=[])

        self.add_retry_handler(404, self._retry_404)

    async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
        if failed_attempts > 1:
            return False
        else:
            log_requests.debug('after 404, about to retry with fresh config')
            self.worker_config = None
            return True

    async def _disconnect(self, timeout: float | None = None):
        client_timeout = None
        if timeout is not None:
            client_timeout = aiohttp.ClientTimeout(total=timeout)
        if (self.is_dev_mode and self.pop_id == 'transient'
                and self.worker_config is not None and self.worker_config.get('pipeline_id') is not None):
            try:
                base_url = await self.dev_mode_base_url()
                delete_pipeline_url = f'{base_url}/pipelines/{self.worker_config["pipeline_id"]}'
                headers = {}
                authorization_header = await self._authorization_header()
                if authorization_header is not None:
                    headers['Authorization'] = authorization_header
                log_requests.debug('before DELETE %s', delete_pipeline_url)
                await self.client_session.delete(delete_pipeline_url, headers=headers, timeout=client_timeout)
                log_requests.debug('after DELETE %s', delete_pipeline_url)
            except Exception as e:
                log.exception(e, exc_info=True)
            finally:
                del self.worker_config["pipeline_id"]

    async def _reconnect(self):
        if self.worker_config is not None:
            return

        if self.last_fetch_config_error_time is not None and self.last_fetch_config_error_time > time.time() - MIN_CONFIG_RECONNECT_SECS:
            raise self.last_fetch_config_error

        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time > time.time() - MIN_CONFIG_RECONNECT_SECS:
            raise aiohttp.ClientConnectionError()

        if self.pop_id == 'transient':
            config_url = f'{self.eyepop_url}/workers/config'
        else:
            config_url = f'{self.eyepop_url}/pops/{self.pop_id}/config?auto_start={self.auto_start}'
        headers = {}
        authorization_header = await self._authorization_header()
        if authorization_header is not None:
            headers['Authorization'] = authorization_header
        try:
            log_requests.debug('before GET %s', config_url)
            async with self.client_session.get(config_url, headers=headers) as response:
                self.worker_config = await response.json()
                self.last_fetch_config_success_time = time.time()
                self.last_fetch_config_error = None
                self.last_fetch_config_error_time = None
        except aiohttp.ClientResponseError as e:
            if e.status != 401:
                self.last_fetch_config_error = e
                self.last_fetch_config_error_time = time.time()
                raise e
            else:
                log_requests.debug('after GET %s: 401, about to retry with fresh access token', config_url)
                self.token = None
                self.expire_token_time = None
                headers = {}
                authorization_header = await self._authorization_header()
                if authorization_header is not None:
                    headers['Authorization'] = authorization_header
                async with self.client_session.get(config_url, headers=headers) as retried_response:
                    self.worker_config = await retried_response.json()
        except aiohttp.ClientConnectionError as e:
            self.last_fetch_config_error = e
            self.last_fetch_config_error_time = time.time()
            raise e

        self.is_dev_mode = self.pop_id == 'transient' or self.worker_config['status'] != 'active_prod'

        log_requests.debug(f'after GET {config_url}: {self.worker_config}')

        base_url = await self.dev_mode_base_url()

        if self.pop_id == 'transient':
            start_pipeline_url = f'{base_url}/pipelines'
            body = {
                "pop": self.pop.model_dump(),
                "source": {
                    "sourceType": "NONE"
                },
                "idleTimeoutSeconds": 60,
                "logging": ["out_meta"],
                "videoOutput": "no_output",
            }
            if self.dataset_uuid is not None:
                body["datasetUuid"] = self.dataset_uuid

            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            log_requests.debug('before POST %s', start_pipeline_url)
            async with self.client_session.post(start_pipeline_url, headers=headers, json=body) as response:
                response_json = await response.json()
            log_requests.debug('after POST %s', start_pipeline_url)
            self.worker_config['pipeline_id'] = response_json['id']

        if self.is_dev_mode and (self.worker_config['base_url'] is None or self.worker_config['pipeline_id'] is None):
            raise PopNotStartedException(pop_id=self.pop_id)

        if self.is_dev_mode and self.stop_jobs:
            stop_jobs_url = f'{await self.dev_mode_pipeline_base_url()}/source?mode=preempt&processing=sync'
            body = {'sourceType': 'NONE'}
            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            log_requests.debug('before PATCH %s', stop_jobs_url)
            async with self.client_session.patch(stop_jobs_url, headers=headers, json=body) as response:
                pass
            log_requests.debug('after PATCH %s', stop_jobs_url)

        if self.is_dev_mode and self.pop is None:
            # get current pipeline string and store
            get_url = await self.dev_mode_pipeline_base_url()
            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            log_requests.debug('before GET %s', get_url)
            async with self.client_session.get(get_url, headers=headers) as response:
                response_json = await response.json()
                pop_as_dict = response_json.get('pop')
                if pop_as_dict is None:
                    self.pop = None
                else:
                    self.pop = Pop(**pop_as_dict)
            log_requests.debug('after GET %s', get_url)
            log.debug('current pop is %s', self.pop)

        if self.is_dev_mode:
            endpoint = {'base_url': urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/"),
                'pipeline_id': self.worker_config['pipeline_id']}
            self.load_balancer = EndpointLoadBalancer([endpoint])
        else:
            self.load_balancer = EndpointLoadBalancer(self.worker_config['endpoints'])


    async def session(self) -> dict:
        session = await super().session()
        session['popId'] = self.pop_id
        if self.worker_config:
            session['baseUrl'] = self.worker_config['base_url']
            session['pipelineId'] = self.worker_config['pipeline_id']
        return session

    async def get_pop(self) -> Pop | None:
        return self.pop

    async def set_pop(self, pop: Pop):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'set_pop not supported in production mode')
        response = await self.pipeline_patch('pop', content_type='application/json',
                                             data=pop.model_dump_json())
        self.pop = pop
        return response

    async def dev_mode_pipeline_base_url(self):
        if self.is_dev_mode:
            return f'{await self.dev_mode_base_url()}/pipelines/{self.worker_config["pipeline_id"]}'
        else:
            pass

    async def upload(
            self,
            location: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _UploadFileJob(
            location=location,
            video_mode=video_mode,
            component_params=params,
            session=self, on_ready=on_ready,
            callback=self.metrics_collector
        )
        await  self._task_start(job.execute())
        return job

    async def upload_stream(
            self,
            stream: BinaryIO,
            mime_type: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _UploadStreamJob(
            stream=stream,
            mime_type=mime_type,
            video_mode=video_mode,
            component_params=params,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def load_from(
            self,
            location: str,
            params: list[ComponentParams] | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _LoadFromJob(
            location=location,
            component_params=params,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def load_asset(
            self,
            asset_uuid: str,
            params: list[ComponentParams] | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _LoadFromAssetUuidJob(
            asset_uuid=asset_uuid,
            component_params=params,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def dev_mode_base_url(self) -> str:
        if self.worker_config is None:
            await self._reconnect()
        return urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/")

    #
    # Implements: _WorkerClientSession
    #

    async def pipeline_get(self, url_path_and_query: str, accept: str | None = None,
                           timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('GET', url_path_and_query, accept=accept, timeout=timeout)

    async def pipeline_post(self, url_path_and_query: str, accept: str | None = None, open_data: Callable = None,
                            content_type: str | None = None,
                            timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('POST', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_patch(self, url_path_and_query: str, accept: str | None = None, data: Any = None,
                             content_type: str | None = None,
                             timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        def noop():
            pass

        def open_data():
            if isinstance(data, str):
                return StringIO(data)
            else:
                return data

        return await self._pipeline_request_with_retry('PATCH', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_delete(self, url_path_and_query: str,
                              timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('DELETE', url_path_and_query, timeout=timeout)

    async def _pipeline_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                           open_data: Callable = None, content_type: str | None = None,
                                           timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time < time.time() - FORCE_REFRESH_CONFIG_SECS:
            self.worker_config is None

        failed_attempts = 0
        retried_re_auth = False
        retried_re_config = False
        start_time = time.time()
        while time.time() - start_time < MAX_RETRY_TIME_SECS:
            if self.worker_config is None:
                await self._reconnect()

            entry = self.load_balancer.next_entry(MAX_RETRY_TIME_SECS + 1)
            if entry is None:
                if not retried_re_config:
                    # pipeline might have just shut down
                    log_requests.debug('no healthy endpoints, about to retry with fresh config')
                    self.worker_config = None
                    retried_re_config = True
                    continue
                else:
                    raise PopNotReachableException(self.pop_id, self.load_balancer.get_debug_status())

            url = f'{entry.base_url}/pipelines/{entry.pipeline_id}/{url_path_and_query}'

            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            if accept is not None:
                headers['Accept'] = accept
            if content_type is not None:
                headers['Content-Type'] = content_type
            try:
                log_requests.debug('before %s %s', method, url)
                if open_data is not None:
                    with open_data() as data:
                        if isinstance(data, StringIO):
                            response = await self.client_session.request(method, url, headers=headers,
                                                                         data=data.getvalue(), timeout=timeout)
                        else:
                            response = await self.client_session.request(method, url, headers=headers, data=data,
                                                                         timeout=timeout)
                else:
                    response = await self.client_session.request(method, url, headers=headers, timeout=timeout)

                entry.mark_success()

                log_requests.debug('after %s %s status=%d headers=%s', method, url, response.status, response.headers)

                return response
            except aiohttp.ClientResponseError as e:
                if e.status == 404:
                    # in load balanced configuration, we overwrite the standard 404 handler
                    entry.mark_error()
                    log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
                else:
                    failed_attempts += 1
                    if e.status not in self.retry_handlers:
                        log_requests.exception('unexpected error', e)
                        raise e
                    if not await self.retry_handlers[e.status](e.status, failed_attempts):
                        log_requests.exception('unexpected error', e)
                        raise e
            except aiohttp.ClientConnectionError:
                entry.mark_error()
                log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
            except Exception as e:
                log_requests.exception('unexpected error', e)
                raise e

        raise PopNotReachableException(self.pop_id, f"no success after {MAX_RETRY_TIME_SECS} secs")

    async def _worker_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                         data: Any = None, content_type: str | None = None,
                                         timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        url = f'{await self.dev_mode_base_url()}/{url_path_and_query}'
        return await self.request_with_retry(method, url, accept, data, content_type, timeout)



import json
import uuid

import pytest
from aioresponses import aioresponses, CallbackResult

from eyepop import EyePopSdk
from eyepop.worker.worker_types import Pop, ForwardComponent, InferenceComponent, InferenceType
from tests.worker.base_endpoint_test import BaseEndpointTest


class TestEndpointPop(BaseEndpointTest):
    test_pop = Pop(
        components=[ForwardComponent()]
    )
    test_new_pop = Pop(
        components=[InferenceComponent(
            inferenceTypes=[InferenceType.OBJECT_DETECTION],
            modelUuid=uuid.uuid4().hex
        )]
    )

    current_pop: Pop | None = None
    @aioresponses()
    def test_sync_get_pop(self, mock: aioresponses):

        self.setup_base_mock(mock, is_transient=True)
        # authentication
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pipeline(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': self.current_pop}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pipeline)

        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id="transient") as endpoint:
            cur_pop = endpoint.get_pop()
        self.assertEqual(cur_pop, Pop(components=[]))
        self.assertBaseMock(mock, is_transient=True)

    @aioresponses()
    def test_sync_set_pop(self, mock: aioresponses):
        self.setup_base_mock(mock, is_transient=True)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pipeline(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': self.current_pop}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pipeline)
        
        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id="transient") as endpoint:
            def set_pop(url, **kwargs) -> CallbackResult:
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                else:
                    self.current_pop = Pop(**json.loads(kwargs['data']))
                    return CallbackResult(status=204)

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/pop',
                       callback=set_pop)
            
            endpoint.set_pop(self.test_new_pop)
            self.assertEqual(endpoint.get_pop(), self.test_new_pop)
            self.assertEqual(self.current_pop, self.test_new_pop)
        self.assertBaseMock(mock, is_transient=True)


    @aioresponses()
    @pytest.mark.asyncio
    async def test_async_get_pop(self, mock: aioresponses):

        self.setup_base_mock(mock, is_transient=True)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pipeline(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': self.current_pop}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pipeline)

        async with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                            pop_id="transient", is_async=True) as endpoint:
            cur_pop = await endpoint.get_pop()
        self.assertEqual(cur_pop, Pop(components=[]))
        self.assertBaseMock(mock, is_transient=True)

    
    @aioresponses()
    @pytest.mark.asyncio
    async def test_async_set_pop(self, mock: aioresponses):

        self.setup_base_mock(mock, is_transient=True)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_pipeline(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'pop': self.current_pop}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_pipeline)

        async with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                            pop_id="transient", is_async=True) as endpoint:
            def set_pop(url, **kwargs) -> CallbackResult:
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                elif kwargs['headers']['Content-Type'] != 'application/json':
                    return CallbackResult(status=400, reason='unsupported content type')
                else:
                    self.current_pop = Pop(**json.loads(kwargs['data']))
                    return CallbackResult(status=204)

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/pop',
                       callback=set_pop)
            
            await endpoint.set_pop(self.test_new_pop)
            self.assertEqual(await endpoint.get_pop(), self.test_new_pop)
            self.assertEqual(self.current_pop, self.test_new_pop)

        self.assertBaseMock(mock, is_transient=True)




import sys
import asyncio
import logging
import time

from eyepop import EyePopSdk
from eyepop import Job
from eyepop.worker.worker_endpoint import WorkerEndpoint

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.getLogger('eyepop.metrics').setLevel(level=logging.DEBUG)


def upload_photos_sequentially(file_paths: list[str]):
    '''
    Sequential processing of batch uploads - simple but slowest option.
    '''
    with EyePopSdk.workerEndpoint() as endpoint:
        for file_path in file_paths:
            job = endpoint.upload(file_path)
            while job.predict() is not None:
                pass


def upload_photos(file_paths: list[str]):
    '''
    Parallel processing of batch uploads - fast but limited by memory
    '''
    with EyePopSdk.workerEndpoint() as endpoint:
        jobs = []
        for file_path in file_paths:
            jobs.append(endpoint.upload(file_path))
        for job in jobs:
            while job.predict() is not None:
                pass


def upload_photos_threaded(file_paths: list[str]):
    '''
    Parallel processing in separate threads - fast but limited by parallelism doesn't adjust to network response times
    '''
    import concurrent.futures

    def run_upload(e: WorkerEndpoint, file_path: str):
        job = e.upload(file_path)
        while job.predict() is not None:
            pass

    with EyePopSdk.workerEndpoint() as endpoint:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(run_upload, endpoint, file_path) for file_path in file_paths]

            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except Exception as exc:
                    print(f"Exception: {exc}")
                    raise exc


async def async_upload_photos(file_paths: list[str]):
    '''
    Async processing of batch uploads - fast and memory efficient
    '''
    sem = asyncio.Semaphore(0)

    async def on_ready(job: Job):
        nonlocal sem
        try:
            while await job.predict() is not None:
                pass
        except Exception as e:
            logging.exception(e)
        finally:
            sem.release()

    async with EyePopSdk.workerEndpoint(is_async=True, job_queue_length=512) as endpoint:
        n = 0
        for file_path in file_paths:
            await endpoint.upload(file_path, on_ready=on_ready)
            n += 1
        for i in range(n):
            await sem.acquire()


example_image_path = sys.argv[1]
example_image_paths = [example_image_path] * int(sys.argv[2])

t1 = time.time()
upload_photos_sequentially(example_image_paths)
t2 = time.time()
print("%d x photo sync took %.3f seconds\n\n" % (len(example_image_paths), (t2 - t1)))

t1 = time.time()
upload_photos(example_image_paths)
t2 = time.time()
print("%d x photo sync took %.3f seconds\n\n" % (len(example_image_paths), (t2 - t1)))

t1 = time.time()
upload_photos_threaded(example_image_paths)
t2 = time.time()
print("%d x photo threaded took %.3f seconds\n\n" % (len(example_image_paths), (t2 - t1)))

t1 = time.time()
asyncio.run(async_upload_photos(example_image_paths))
t2 = time.time()
print("%d x photo async took %.3f seconds\n\n" % (len(example_image_paths), (t2 - t1)))



import pyarrow as pa
import pandas as pd
from pyarrow import ListType, StructType

from eyepop.data.arrow.schema import ASSET_SCHEMA
from eyepop.data.arrow.schema_version_conversion import convert


def _pandas_compatible_data_type(data_type: pa.DataType) -> pa.DataType:
    if data_type == pa.float16():
        return pa.float64()
    elif isinstance(data_type, ListType):
        return pa.list_(_pandas_compatible_data_type(data_type.field(0).type))
    elif isinstance(data_type, StructType):
        pandas_compatible_fields = [None] * data_type.num_fields
        for i in range(data_type.num_fields):
            field = data_type.field(i)
            pandas_compatible_field = pa.field(field.name, _pandas_compatible_data_type(field.type))
            pandas_compatible_fields[i] = pandas_compatible_field
        return pa.struct(pandas_compatible_fields)
    else:
        return data_type

def _pandas_compatible_schema(schema: pa.Schema) -> pa.Schema:
    pandas_compatible_fields = [None] * len(schema)
    for i in range(len(schema)):
        field = schema.field(i)
        pandas_compatible_fields[i] = pa.field(field.name, _pandas_compatible_data_type(field.type))
    return pa.schema(fields=pandas_compatible_fields)

PANDAS_COMPAT_SCHEMA = _pandas_compatible_schema(ASSET_SCHEMA)

def table_from_dataframe(df: pd.DataFrame, schema: pa.Schema = ASSET_SCHEMA) -> pa.Table:
    """ Safely converts a pandas dataframe to a pyarrow.Table that complies with the ASSET_SCHEMA.
    """
    compat_schema = PANDAS_COMPAT_SCHEMA
    if schema != ASSET_SCHEMA:
        compat_schema = _pandas_compatible_schema(schema)
    unsafe_table = pa.Table.from_pandas(df, schema=compat_schema)
    return unsafe_table.cast(schema)


def dataframe_from_table(table: pa.Table, schema: pa.Schema = ASSET_SCHEMA) -> pd.DataFrame:
    """ Safely converts a pyarrow.Table with ASSET_SCHEMA to a pandas dataframe.
    """
    type_mapping = {
        # Needed that to avoid exception when categorical column is completely None
        pa.dictionary(pa.int32(), pa.string()): pd.StringDtype(),
    }
    return convert(table, schema).to_pandas(types_mapper=type_mapping.get)




import pyarrow as pa
from eyepop.data.data_types import AssetAnnotationResponse, Prediction

from eyepop.data.arrow.eyepop.predictions import table_from_eyepop_predicted_objects, \
    eyepop_predicted_objects_from_pylist, \
    eyepop_predicted_classes_from_pylist, table_from_eyepop_predicted_classes, table_from_eyepop_predicted_key_pointss, \
    eyepop_predicted_key_pointss_from_pylist, table_from_eyepop_predicted_texts, eyepop_predicted_texts_from_pylist, \
    eyepop_predicted_embeddings_from_pylist, table_from_eyepop_predicted_embeddings
from eyepop.data.arrow.schema import ANNOTATION_SCHEMA
from eyepop.data.arrow.schema_version_conversion import convert


def table_from_eyepop_annotations(annotations: list[AssetAnnotationResponse], schema: pa.Schema = ANNOTATION_SCHEMA) -> pa.Table:
    types = []
    sources = []
    user_reviews = []
    objects = []
    classes = []
    source_model_uuids = [] if "source_model_uuid" in schema.names else None
    key_points = [] if "keyPoints" in schema.names else None
    texts = [] if "texts" in schema.names else None
    embeddings = [] if "embeddings" in schema.names else None
    timestamps = [] if "timestamp" in schema.names else None
    durations = [] if "duration" in schema.names else None
    offsets = [] if "offset" in schema.names else None
    offset_durations = [] if "offset_duration" in schema.names else None

    for e in annotations:
        types.append(e.type)
        sources.append(e.source)
        user_reviews.append(e.user_review)
        if e.annotation.objects is None:
            objects.append(None)
        elif len(e.annotation.objects) == 0:
            objects.append([])
        else:
            objects.append(table_from_eyepop_predicted_objects(
                e.annotation.objects,
                e.annotation.source_width,
                e.annotation.source_height,
                e.user_review,
                schema=pa.schema(schema.field(3).type.value_type), # schema for "objects" field
            ).to_struct_array())
        if e.annotation.classes is None:
            classes.append(None)
        elif len(e.annotation.classes) == 0:
            classes.append([])
        else:
            classes.append(table_from_eyepop_predicted_classes(
                e.annotation.classes,
                e.user_review,
                schema=pa.schema(schema.field(4).type.value_type),  # schema for "classes" field
            ).to_struct_array())
        if source_model_uuids is not None:
            source_model_uuids.append(e.source_model_uuid)
        if key_points is not None:
            if e.annotation.keyPoints is None:
                key_points.append(None)
            elif len(e.annotation.keyPoints) == 0:
                key_points.append([])
            else:
                key_points.append(table_from_eyepop_predicted_key_pointss(
                    e.annotation.keyPoints,
                    e.annotation.source_width,
                    e.annotation.source_height,
                    schema=pa.schema(schema.field(6).type.value_type),  # schema for "keyPoints" field
                ).to_struct_array())
        if texts is not None:
            if e.annotation.texts is None:
                texts.append(None)
            elif len(e.annotation.texts) == 0:
                texts.append([])
            else:
                texts.append(table_from_eyepop_predicted_texts(e.annotation.texts).to_struct_array())
        if embeddings is not None:
            if e.annotation.embeddings is None:
                embeddings.append(None)
            elif len(e.annotation.embeddings) == 0:
                embeddings.append([])
            else:
                embeddings.append(table_from_eyepop_predicted_embeddings(
                    e.annotation.embeddings,
                    schema=pa.schema(schema.field(8).type.value_type),  # schema for "embeddings" field
                ).to_struct_array())
        if timestamps is not None:
            timestamps.append(e.annotation.timestamp)
        if durations is not None:
            durations.append(e.annotation.duration)
        if offsets is not None:
            offsets.append(e.annotation.offset)
        if offset_durations is not None:
            offset_durations.append(e.annotation.offset_duration)

    # since 0.0
    columns = [
        pa.array(types).dictionary_encode(),
        pa.array(sources).dictionary_encode(),
        pa.array(user_reviews).dictionary_encode(),
        objects,
        classes,
    ]

    # since 1.0: source_model_uuid
    if source_model_uuids is not None:
        columns.append(pa.array(source_model_uuids).dictionary_encode())

    # since 1.1: key_points
    if key_points is not None:
        columns.append(key_points)

    # since 1.2: texts
    if texts is not None:
        columns.append(texts)

    # since 1.3: embeddings
    if embeddings is not None:
        columns.append(embeddings)

    # since 1.4: timestamp, duration, offset, offset_duration
    if timestamps is not None:
        columns.append(timestamps)
    if durations is not None:
        columns.append(durations)
    if offsets is not None:
        columns.append(offsets)
    if offset_durations is not None:
        columns.append(offset_durations)
    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_annotations_from_table(table: pa.Table) -> list[AssetAnnotationResponse]:
    table = convert(table, ANNOTATION_SCHEMA)
    annotations = []
    i = 0
    for batch in table.to_reader():
        types = batch.column(0).to_pylist()
        sources = batch.column(1).to_pylist()
        user_reviews = batch.column(2).to_pylist()
        objects = batch.column(3).to_pylist()
        classes = batch.column(4).to_pylist()
        # since 1.0: source_model_uuid
        source_model_uuid = batch.column(5).to_pylist()
        # since 1.1: key_points
        key_points = batch.column(6).to_pylist()
        # since 1.2: texts
        texts = batch.column(7).to_pylist()
        # since 1.3: embeddings
        embeddings = batch.column(8).to_pylist()
        # since 1.4: timestamp, duration, offset, offset_duration
        timestamps = batch.column(9).to_pylist()
        durations = batch.column(10).to_pylist()
        offsets = batch.column(11).to_pylist()
        offset_durations = batch.column(12).to_pylist()

        for j in range(len(types)):
            if objects[j] is None:
                child_objects = None
            elif len(objects[j]) == 0:
                child_objects = []
            else:
                child_objects = eyepop_predicted_objects_from_pylist(objects[j], 1.0, 1.0)

            if classes[j] is None:
                child_classes = None
            elif len(classes[j]) == 0:
                child_classes = []
            else:
                child_classes = eyepop_predicted_classes_from_pylist(classes[j])

            if key_points[j] is None:
                child_key_pointss = None
            elif len(key_points[j]) == 0:
                child_key_pointss = []
            else:
                child_key_pointss = eyepop_predicted_key_pointss_from_pylist(key_points[j], 1.0, 1.0)

            if texts[j] is None:
                child_texts = None
            elif len(texts[j]) == 0:
                child_texts = []
            else:
                child_texts = eyepop_predicted_texts_from_pylist(texts[j])

            if embeddings[j] is None:
                child_embeddings = None
            elif len(embeddings[j]) == 0:
                child_embeddings = []
            else:
                child_embeddings = eyepop_predicted_embeddings_from_pylist(embeddings[j])

            annotations.append(AssetAnnotationResponse(
                type=types[j],
                user_review=user_reviews[j],
                source=sources[j],
                annotation=Prediction(
                    source_width=1.0,
                    source_height=1.0,
                    timestamp=timestamps[j],
                    duration=durations[j],
                    offset=offsets[j],
                    offset_duration=offset_durations[j],
                    objects=child_objects,
                    classes=child_classes,
                    keyPoints=child_key_pointss,
                    texts=child_texts,
                    embeddings=child_embeddings,
                ),
                source_model_uuid=source_model_uuid[j],
            ))
            i += 1
    return annotations


def eyepop_annotations_from_pylist(py_list: list[dict]) -> list[AssetAnnotationResponse]:
    annotations = []
    for i, o in enumerate(py_list):
        if o is None:
            continue
        objects = o.get('objects', None)
        if objects is None:
            child_objects = None
        else:
            child_objects = eyepop_predicted_objects_from_pylist(objects, 1.0, 1.0)
        classes = o.get('classes', None)
        if classes is None:
            child_classes = None
        else:
            child_classes = eyepop_predicted_classes_from_pylist(classes)
        key_pointss = o.get("keyPoints", None)
        if key_pointss is None:
            child_key_pointss = None
        else:
            child_key_pointss = eyepop_predicted_key_pointss_from_pylist(key_pointss, 1.0, 1.0)
        texts = o.get("texts", None)
        if texts is None:
            child_texts = None
        else:
            child_texts = eyepop_predicted_texts_from_pylist(texts)
        embeddings = o.get("embeddings", None)
        if embeddings is None:
            child_embeddings = None
        else:
            child_embeddings = eyepop_predicted_embeddings_from_pylist(embeddings)
        annotations.append(AssetAnnotationResponse(
            type=o['type'],
            user_review=o['user_review'],
            source=o['source'],
            annotation=Prediction(
                source_width=1.0,
                source_height=1.0,
                timestamp=o.get('timestamp', None),
                duration=o.get('duration', None),
                offset=o.get('offset', None),
                offset_duration=o.get('offset_duration', None),
                objects=child_objects,
                classes=child_classes,
                keyPoints=child_key_pointss,
                texts=child_texts,
                embeddings=child_embeddings
            ),
            source_model_uuid = o.get('source_model_uuid', None)
        ))
    return annotations



import argparse
import asyncio
import json
import logging
import os
import sys
from typing import Any

from eyepop import EyePopSdk, Job
from eyepop.worker.worker_types import Pop, InferenceComponent, FullForward

script_dir = os.path.dirname(__file__)

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.getLogger('eyepop.requests').setLevel(level=logging.INFO)

log = logging.getLogger('eyepop.example')

parser = argparse.ArgumentParser(
                    prog='Pop examples',
                    description='Demonstrates the caption generation via VLM/LLM',
                    epilog='.')
parser.add_argument('-l', '--local-path', required=False, type=str, default=False, help="run the inference on a local file or directory")
parser.add_argument('-a', '--asset-uuid', required=False, type=str, default=False, help="run the inference on an asset by its Uuid")
parser.add_argument('-u', '--url', required=False, type=str, default=False, help="run the inference on a remote Url")
parser.add_argument('-c', '--caption-ability', type=str, default="eyepop.vlm.preview:latest", help="run this ability to generate captions", choices=[
    "eyepop.vlm.preview:latest",
    "eyepop.image-caption.preview.qwen2:latest",
    "eyepop.image-caption.preview.qwen2-float16:latest"
])
parser.add_argument('-r', '--role', type=str, default=None, required=False, help="use this role to generate the caption")
parser.add_argument('-p', '--prompt', type=str, default=None, required=False, help="use this prompt to generate the caption")
parser.add_argument('-q', '--question', type=str, default=None, required=False, help="apply this question to the caption")


main_args = parser.parse_args()

if not main_args.local_path and not main_args.url and not main_args.asset_uuid:
    print("Need something to run inference on; pass either --url or --local-path or --asset-uuid")
    parser.print_help()
    sys.exit(1)

forward = None
if main_args.question is not None:
    forward = FullForward(
        targets=[InferenceComponent(
            id=2,
            categoryName='answers',
            model='eyepop.question-answer.preview:latest',
            params={
                "prompt": main_args.question
            }
        )]
    )

caption_params = dict()
if main_args.role is not None:
    caption_params['role'] = main_args.role
if main_args.prompt is not None:
    caption_params['prompt'] = main_args.prompt

caption_pop = Pop(components=[
    InferenceComponent(
        id=1,
        categoryName='captions',
        ability=main_args.caption_ability,
        forward=forward,
        params=caption_params,
    )
])

params = None

async def main(args):
    async with EyePopSdk.workerEndpoint(is_async=True) as endpoint:
        await endpoint.set_pop(caption_pop)
        if args.local_path:
            if not os.path.exists(args.local_path):
                log.warning(f"local path {args.local_path} does not exist")
                sys.exit(1)
            if os.path.isfile(args.local_path):
                local_files = [args.local_path]
            else:
                local_files = []
                for f in os.listdir(args.local_path):
                    local_file = os.path.join(args.local_path, f)
                    if os.path.isfile(local_file):
                        local_files.append(local_file)
            jobs = []
            async def on_ready(job: Job, path: str):
                while result := await job.predict():
                    print(path, json.dumps(result, indent=2))
            for local_file in local_files:
                job = await endpoint.upload(local_file, params=params)
                await on_ready(job, local_file)
            #     jobs.append(on_ready(job, local_file))
            # await asyncio.gather(*jobs)
        elif args.url:
            job = await endpoint.load_from(args.url, params=params)
            while result := await job.predict():
                print(json.dumps(result, indent=2))
        elif args.asset_uuid:
            job = await endpoint.load_asset(args.asset_uuid, params=params)
            while result := await job.predict():
                print(json.dumps(result, indent=2))

asyncio.run(main(main_args))


from eyepop.data.data_types import Prediction, PredictedClass

import heapq

def _top_k(classes: list[PredictedClass], k: int) -> list[PredictedClass]:
    """ Find k classes with the top confidence """

    if len(classes) <= k:
        return classes

    # extract (comparison value, value) as heapifyalbe items, using i asa tiebreaker if confidences are equal
    classes = [((clazz.confidence if clazz.confidence is not None else -1.0), i, clazz) for i, clazz in enumerate(classes)]

    # Create a min-heap with the first k elements
    min_heap = classes[:k]
    heapq.heapify(min_heap)

    # Traverse the rest of the array
    for x in classes[k:]:
        if x[0] > min_heap[0][0]:
            heapq.heappushpop(min_heap, x)

    res = []

    # Min heap will contain only k
    # largest element
    while min_heap:
        res.append(heapq.heappop(min_heap))

    # Reverse the result array, so that all
    # elements are in decreasing order
    res.reverse()

    return [i[2] for i in res]

def filter_prediction_top_k(prediction: Prediction, k: int) -> Prediction:
    objects = prediction.objects
    classes = prediction.classes
    if (objects is None or len(objects) <= k) and (classes is None or len(classes) <= k):
        return prediction
    return Prediction(
        source_width=prediction.source_width,
        source_height=prediction.source_height,
        objects=_top_k(objects, k) if objects is not None else None,
        classes=_top_k(classes, k) if classes is not None else None,
        texts=prediction.texts,
        meshs=prediction.meshs,
        keyPoints=prediction.keyPoints,
    )



import asyncio
import json
from asyncio import StreamReader
from typing import Callable, BinaryIO, Any, AsyncIterable
from urllib.parse import urljoin, quote_plus

import aiohttp
import websockets
from pydantic import TypeAdapter
from websockets.asyncio.client import ClientConnection

from pydantic.tools import parse_obj_as

from eyepop.client_session import ClientSession
from eyepop.data.arrow.schema import MIME_TYPE_APACHE_ARROW_FILE_VERSIONED
from eyepop.data.data_jobs import DataJob, _UploadStreamJob, _ImportFromJob
from eyepop.data.data_syncify import SyncDataJob
from eyepop.data.data_types import ArgoWorkflowPhase, Dataset, DatasetCreate, DatasetUpdate, Asset, ListWorkflowItem, Prediction, \
    AssetImport, AutoAnnotate, UserReview, TranscodeMode, Model, ModelCreate, ModelUpdate, \
    ModelTrainingProgress, ChangeEvent, ChangeType, EventHandler, ModelAlias, ModelAliasCreate, \
    ModelAliasUpdate, ModelExportFormat, QcAiHubExportParams, AssetUrlType, AssetInclusionMode, AnnotationInclusionMode, \
    ModelTrainingAuditRecord, ExportedUrlResponse, ModelTrainingEvent, ArtifactType, CreateWorkflowBody, CreateWorkflowResponse
from eyepop.endpoint import Endpoint, log_requests

APPLICATION_JSON = "application/json"

WS_INITIAL_RECONNECT_DELAY = 1.0
WS_MAX_RECONNECT_DELAY = 60.0


class DataClientSession(ClientSession):
    def __init__(self, delegee: ClientSession, base_url: str):
        self.delegee = delegee
        self.base_url = base_url

    async def request_with_retry(self, method: str, url: str, accept: str | None = None, data: Any = None,
                                 content_type: str | None = None,
                                 timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        url = urljoin(self.base_url, url)
        return await self.delegee.request_with_retry(method, url, accept, data, content_type, timeout)


class DataEndpoint(Endpoint):
    """
    Endpoint to the EyePop.ai Data API.
    """
    account_uuid: str
    data_config: dict[str, Any] | None

    disable_ws: bool
    ws: ClientConnection | None
    ws_tasks = set[asyncio.Task]
    ws_current_reconnect_delay: float | None
    account_event_handlers: set[EventHandler]
    dataset_uuid_to_event_handlers: dict[str, set[EventHandler]]

    def __init__(self, secret_key: str | None, access_token: str | None,
                 eyepop_url: str, account_id: str, job_queue_length: int,
                 request_tracer_max_buffer: int, disable_ws: bool = True):
        super().__init__(
            secret_key=secret_key, access_token=access_token, eyepop_url=eyepop_url,
            job_queue_length=job_queue_length, request_tracer_max_buffer=request_tracer_max_buffer
        )
        self.account_uuid = account_id
        self.data_config = None

        self.disable_ws = disable_ws
        self.ws = None
        self.ws_tasks = set()
        self.ws_current_reconnect_delay = None
        self.account_event_handlers = set()
        self.dataset_uuid_to_event_handlers = dict()

        self.add_retry_handler(404, self._retry_404)

    async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
        if failed_attempts > 1:
            return False
        else:
            log_requests.debug('after 404, about to retry with fresh config')
            self.data_config = None
            return True

    async def _disconnect(self, timeout: float | None = None):
        await self._ws_disconnect()

    async def _reconnect(self):
        if self.data_config is not None:
            return
        config_url = f'{self.eyepop_url}/data/config?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", config_url) as resp:
            self.data_config = await resp.json()
        if not self.disable_ws:
            await self._reconnect_ws()

    async def _reconnect_ws(self):
        await self._ws_disconnect()
        ws_url = urljoin(urljoin(
            self.eyepop_url, self.data_config['base_url']
        ).rstrip('/'),'events').replace(
            "https://", "wss://"
        ).replace(
            "http://", "ws://"
        )

        log_requests.debug("before ws connect: %s", ws_url)
        ws = await websockets.asyncio.client.connect(uri=ws_url)
        log_requests.debug("after ws connect: %s", ws_url)
        authorization_header = await self._authorization_header()
        if authorization_header is not None:
            auth_headers = {'authorization': authorization_header}
            message = json.dumps(auth_headers)
            await ws.send(message)
            log_requests.debug("ws send: %s", message)
        message = json.dumps({
            "subscribe" : {
                "account_uuid": self.account_uuid
            }
        })
        await ws.send(message)
        log_requests.debug("ws send: %s", message)
        for dataset_uuid in self.dataset_uuid_to_event_handlers:
            message = json.dumps({
                "subscribe" : {
                    "dataset_uuid": dataset_uuid
                }
            })
            await ws.send(message)
            log_requests.debug("ws send: %s", message)

        self.ws_current_reconnect_delay = WS_INITIAL_RECONNECT_DELAY
        ws_reader_task = asyncio.create_task(self._ws_reader(ws))

        self.ws = ws
        self.ws_tasks.add(ws_reader_task)
        ws_reader_task.add_done_callback(self.ws_tasks.discard)

    async def _ws_disconnect(self):
        ws = self.ws
        self.ws = None
        self.ws_reader_task = None
        if ws is not None:
            await ws.close()

    async def _ws_reader(self, ws: ClientConnection):
        try:
            async for message in ws:
                log_requests.debug("ws received %s", message)
                try:
                    data = json.loads(message)
                    if "change_type" in data:
                        change_event = ChangeEvent(**data)
                        await self._dispatch_change_event(change_event)
                except Exception as e:
                    log_requests.exception(e)
        except websockets.ConnectionClosed:
            log_requests.debug("ws disconnected")
            await self._ws_disconnect()
            if self.ws_current_reconnect_delay is None:
                self.ws_current_reconnect_delay = WS_INITIAL_RECONNECT_DELAY
            elif self.ws_current_reconnect_delay < WS_MAX_RECONNECT_DELAY:
                self.ws_current_reconnect_delay *= 1.5
            await asyncio.sleep(self.ws_current_reconnect_delay)
            ws_reconnect_task = asyncio.create_task(self._reconnect_ws())
            self.ws_tasks.add(ws_reconnect_task)
            ws_reconnect_task.add_done_callback(self.ws_tasks.discard)

    account_event_types = {
        ChangeType.dataset_added,
        ChangeType.dataset_modified,
        ChangeType.dataset_removed,
        ChangeType.dataset_version_modified,
        ChangeType.workflow_started,
        ChangeType.workflow_succeeded,
        ChangeType.workflow_failed,
        ChangeType.workflow_task_started,
        ChangeType.workflow_task_succeeded,
        ChangeType.workflow_task_failed

    }
    dataset_event_handlers = {
        ChangeType.dataset_added,
        ChangeType.dataset_modified,
        ChangeType.dataset_removed,
        ChangeType.dataset_version_modified,
        ChangeType.asset_added,
        ChangeType.asset_removed,
        ChangeType.asset_status_modified,
        ChangeType.asset_annotation_modified,
        ChangeType.model_added,
        ChangeType.model_modified,
        ChangeType.model_removed,
        ChangeType.model_status_modified,
        ChangeType.model_progress
    }

    async def _dispatch_change_event(self, change_event: ChangeEvent) -> None:
        if change_event.change_type in self.account_event_handlers:
            event_handlers = self.account_event_handlers.copy()
            for handler in event_handlers:
                await handler(change_event)
        if change_event.change_type in self.dataset_event_handlers:
            event_handlers = self.dataset_uuid_to_event_handlers.get(change_event.dataset_uuid, None)
            if event_handlers is not None:
                event_handlers = event_handlers.copy()
                for handler in event_handlers:
                    await handler(change_event)

    async def data_base_url(self) -> str:
        if self.data_config is None:
            await self._reconnect()
        return urljoin(self.eyepop_url, self.data_config['base_url']).rstrip("/")

    """ Event handlers """
    async def add_account_event_handler(self, event_handler: EventHandler):
        if self.disable_ws:
            raise ValueError("event handlers disabled, create endpoint with disable_ws=False "
                             "to register event handlers")
        self.account_event_handlers.add(event_handler)

    async def remove_account_event_handler(self, event_handler: EventHandler):
        self.account_event_handlers.discard(event_handler)

    async def add_dataset_event_handler(self, dataset_uuid: str, event_handler: EventHandler):
        if self.disable_ws:
            raise ValueError("event handlers disabled, create endpoint with disable_ws=False "
                             "to register event handlers")
        event_handlers = self.dataset_uuid_to_event_handlers.get(dataset_uuid, None)
        if event_handlers is None:
            event_handlers = set()
            self.dataset_uuid_to_event_handlers[dataset_uuid] = event_handlers
            ws = self.ws
            if ws:
                message = json.dumps({
                    "subscribe": {
                        "dataset_uuid": dataset_uuid
                    }
                })
                await ws.send(message)
                log_requests.debug("ws send: %s", message)

        event_handlers.add(event_handler)

    async def remove_dataset_event_handler(self, dataset_uuid: str, event_handler: EventHandler):
        event_handlers = self.dataset_uuid_to_event_handlers.get(dataset_uuid, None)
        if event_handlers is not None:
            event_handlers.discard(event_handler)
            if len(event_handlers) == 0:
                del self.dataset_uuid_to_event_handlers[dataset_uuid]
                ws = self.ws
                if ws:
                    message = json.dumps({
                        "unsubscribe": {
                            "dataset_uuid": dataset_uuid
                        }
                    })
                    await ws.send(message)
                    log_requests.debug("ws send: %s", message)

    async def remove_all_dataset_event_handlers(self, dataset_uuid: str):
        event_handlers = self.dataset_uuid_to_event_handlers.get(dataset_uuid, None)
        if event_handlers is not None:
            del self.dataset_uuid_to_event_handlers[dataset_uuid]
            ws = self.ws
            if ws:
                message = json.dumps({
                    "unsubscribe": {
                        "dataset_uuid": dataset_uuid
                    }
                })
                await ws.send(message)
                log_requests.debug("ws send: %s", message)

    """ Model methods """

    async def list_datasets(
            self,
            include_hero_asset: bool = False,
            modifiable_version_only: bool | None = None
    ) -> list[Dataset]:
        modifiable_version_only_query = f'&modifiable_version_only={modifiable_version_only}' if modifiable_version_only is not None else ''
        get_url = f'{await self.data_base_url()}/datasets?account_uuid={self.account_uuid}&include_hero_asset={include_hero_asset}{modifiable_version_only_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return TypeAdapter(list[Dataset]).validate_python(await resp.json())

    async def create_dataset(self, dataset: DatasetCreate) -> Dataset:
        post_url = f'{await self.data_base_url()}/datasets?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=dataset.model_dump_json(exclude_unset=True)) as resp:
            return TypeAdapter(Dataset).validate_python(await resp.json())

    async def get_dataset(
            self,
            dataset_uuid: str,
            dataset_version: int | None = None,
            include_stats: bool = False,
            modifiable_version_only: bool | None = None
    ) -> Dataset:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        modifiable_version_only_query = f'&modifiable_version_only={modifiable_version_only}' if modifiable_version_only is not None else ''
        get_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}?include_stats={include_stats}{version_query}{modifiable_version_only_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return TypeAdapter(Dataset).validate_python(await resp.json())

    async def update_dataset(self, dataset_uuid: str, dataset: DatasetUpdate, start_auto_annotate: bool = True) -> Dataset:
        patch_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}?start_auto_annotate={start_auto_annotate}'
        log_requests.debug('update_dataset: %s', dataset.model_dump_json())
        async with await self.request_with_retry("PATCH", patch_url, content_type=APPLICATION_JSON,
                                                 data=dataset.model_dump_json(exclude_unset=True, exclude_none=True)) as resp:
            return TypeAdapter(Dataset).validate_python(await resp.json())

    async def delete_dataset(self, dataset_uuid: str) -> None:
        delete_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}'
        async with await self.request_with_retry("DELETE", delete_url):
            return

    async def analyze_dataset_version(self, dataset_uuid: str, dataset_version: int | None = None) -> None:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}/analyze?{version_query}'
        async with await self.request_with_retry("POST", post_url):
            return

    async def auto_annotate_dataset_version(self, dataset_uuid: str, dataset_version: int | None = None, max_assets: int | None = None) -> None:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        max_assets_query = f'&max_assets={max_assets}' if max_assets is not None else ''
        post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}/auto_annotate?{version_query}{max_assets_query}'
        async with await self.request_with_retry("POST", post_url):
            return

    async def freeze_dataset_version(self, dataset_uuid: str, dataset_version: int | None = None) -> Dataset:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}/freeze?{version_query}'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(Dataset, await resp.json())

    async def delete_dataset_version(self, dataset_uuid: str, dataset_version: int) -> Dataset:
        delete_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}/versions?dataset_version={dataset_version}'
        async with await self.request_with_retry("DELETE", delete_url) as resp:
            return parse_obj_as(Dataset, await resp.json())

    async def delete_annotations(self, dataset_uuid: str, dataset_version: int,
                                 user_reviews: list[UserReview] = (UserReview.unknown,)) -> None:
        user_reviews_query = ""
        for user_review in user_reviews:
            user_reviews_query += f"&user_review={user_review}"
        delete_url = (f'{await self.data_base_url()}/datasets/{dataset_uuid}/annotations'
                    f'?dataset_version={dataset_version}{user_reviews_query}')
        async with await self.request_with_retry("DELETE", delete_url):
            return

    """" Asset methods """

    async def upload_asset_job(
            self,
            stream: BinaryIO,
            mime_type: str,
            dataset_uuid: str,
            dataset_version: int | None = None,
            external_id: str | None = None,
            sync_transform: bool | None = None,
            no_transform: bool | None = None,
            on_ready: Callable[[DataJob], None] | None = None
    ) -> DataJob | SyncDataJob:
        session = DataClientSession(self, await self.data_base_url())
        job = _UploadStreamJob(
            stream=stream,
            mime_type=mime_type,
            dataset_uuid=dataset_uuid,
            dataset_version=dataset_version,
            external_id=external_id,
            sync_transform=sync_transform,
            no_transform=no_transform,
            session=session,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def import_asset_job(
            self,
            asset_import: AssetImport,
            dataset_uuid: str,
            dataset_version: int | None = None,
            external_id: str | None = None,
            partition: str | None = None,
            sync_transform: bool | None = None,
            no_transform: bool | None = None,
            on_ready: Callable[[DataJob], None] | None = None
    ) -> DataJob | SyncDataJob:
        session = DataClientSession(self, await self.data_base_url())
        job = _ImportFromJob(
            asset_import=asset_import,
            dataset_uuid=dataset_uuid,
            dataset_version=dataset_version,
            external_id=external_id,
            sync_transform=sync_transform,
            no_transform=no_transform,
            session=session,
            partition=partition,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def list_assets(self, dataset_uuid: str, dataset_version: int | None = None,
                          include_annotations: bool = False) -> list[Asset]:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets?dataset_uuid={dataset_uuid}&include_annotations={"true" if include_annotations else "false"}{version_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(list[Asset], await resp.json())

    async def get_asset(self, asset_uuid: str, dataset_uuid: str | None = None,
                        dataset_version: int | None = None, include_annotations: bool = False) -> Asset:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets/{asset_uuid}?include_annotations={"true" if include_annotations else "false"}{dataset_query}{version_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(Asset, await resp.json())

    async def delete_asset(self, asset_uuid: str, dataset_uuid: str | None = None,
                           dataset_version: int | None = None) -> None:
        dataset_query = f'dataset_uuid={dataset_uuid}&' if dataset_uuid is not None else ''
        version_query = f'dataset_version={dataset_version}' if dataset_version is not None else ''
        delete_url = f'{await self.data_base_url()}/assets/{asset_uuid}?{dataset_query}{version_query}'
        async with await self.request_with_retry("DELETE", delete_url):
            return

    async def resurrect_asset(self, asset_uuid: str, dataset_uuid: str, from_dataset_version: int,
                              into_dataset_version: int | None = None) -> None:
        into_version_query = f'&into_dataset_version={into_dataset_version}' if into_dataset_version is not None else ''
        post_url = f'{await self.data_base_url()}/assets/{asset_uuid}/resurrect?dataset_uuid={dataset_uuid}&from_dataset_version={from_dataset_version}{into_version_query}'
        async with await self.request_with_retry("POST", post_url):
            return

    async def update_asset_ground_truth(self, asset_uuid: str, dataset_uuid: str | None = None,
                                        dataset_version: int | None = None,
                                        ground_truth: Prediction | None = None) -> None:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        patch_url = f'{await self.data_base_url()}/assets/{asset_uuid}/ground_truth?{dataset_query}{version_query}'
        async with await self.request_with_retry("PATCH", patch_url,
                                                 content_type=APPLICATION_JSON if ground_truth else None,
                                                 data=ground_truth.model_dump_json(
                                                     exclude_unset=True, exclude_none=True
                                                 ) if ground_truth else None):
            return

    async def delete_asset_ground_truth(self, asset_uuid: str, dataset_uuid: str | None = None,
                                        dataset_version: int | None = None) -> None:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        patch_url = f'{await self.data_base_url()}/assets/{asset_uuid}/ground_truth?{dataset_query}{version_query}'
        async with await self.request_with_retry("DELETE", patch_url):
            return

    async def update_asset_auto_annotation_status(self, asset_uuid: str, auto_annotate: AutoAnnotate,
                                                  user_review: UserReview, approved_threshold: float | None = None,
                                                  dataset_uuid: str | None = None,
                                                  dataset_version: int | None = None) -> None:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        threshold_query = f'&approved_threshold={approved_threshold}' if approved_threshold is not None else ''
        patch_url = (f'{await self.data_base_url()}/assets/{asset_uuid}/auto_annotations/{auto_annotate}/user_review/'
                     f'{user_review}?{dataset_query}{version_query}{threshold_query}')
        async with await self.request_with_retry("PATCH", patch_url):
            return

    async def download_asset(self, asset_uuid: str, dataset_uuid: str | None = None,
                             dataset_version: int | None = None,
                             transcode_mode: TranscodeMode = TranscodeMode.original) -> StreamReader:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets/{asset_uuid}/download?transcode_mode={transcode_mode}{dataset_query}{version_query}'
        resp = await self.request_with_retry("GET", get_url)
        return resp.content

    """ Model methods """

    async def list_models(self) -> list[Model]:
        get_url = f'{await self.data_base_url()}/models?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(list[Model], await resp.json())

    async def create_model(self, model: ModelCreate) -> Model:
        post_url = f'{await self.data_base_url()}/models?account_uuid={self.account_uuid}&start_training=False'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=model.model_dump_json(exclude_unset=True)) as resp:
            return parse_obj_as(Model, await resp.json())

    async def upload_model_artifact(self, model_uuid: str, model_format: ModelExportFormat, artifact_name: str,
                                    stream: BinaryIO, mime_type: str = 'application/octet-stream') -> None:
        put_url = f'{await self.data_base_url()}/models/{model_uuid}/exports/eyepop/formats/{model_format}/artifacts/{artifact_name}'
        async with await self.request_with_retry("PUT", put_url, data=stream, content_type=mime_type,
                                                 timeout=aiohttp.ClientTimeout(total=None, sock_read=60)):
            return

    async def create_model_from_dataset(self, dataset_uuid: str, dataset_version: int | None, model: ModelCreate, start_training: bool = True) -> Model:
        post_url = f'{await self.data_base_url()}/models?dataset_uuid={dataset_uuid}&dataset_version={dataset_version}&start_training={start_training}'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=model.model_dump_json(exclude_unset=True)) as resp:
            return parse_obj_as(Model, await resp.json())

    async def get_model(self, model_uuid: str) -> Model:
        get_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(Model, await resp.json())

    async def get_model_progress(self, model_uuid: str) -> ModelTrainingProgress:
        get_url = f'{await self.data_base_url()}/models/{model_uuid}/progress'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(ModelTrainingProgress, await resp.json())

    async def update_model(self, model_uuid: str, model: ModelUpdate) -> Model:
        patch_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("PATCH", patch_url, content_type=APPLICATION_JSON,
                                                 data=model.model_dump_json(
                                                     exclude_unset=True, exclude_none=True
                                                 )) as resp:
            return parse_obj_as(Model, await resp.json())

    async def delete_model(self, model_uuid: str) -> None:
        delete_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("DELETE", delete_url):
            return

    async def train_model(self, model_uuid: str) -> Model:
        post_url = f'{await self.data_base_url()}/models/{model_uuid}/train'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(Model, await resp.json())

    async def publish_model(self, model_uuid: str) -> Model:
        post_url = f'{await self.data_base_url()}/models/{model_uuid}/publish'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(Model, await resp.json())

    """ Model aliases methods """

    async def list_model_aliases(self) -> list[ModelAlias]:
        get_url = f'{await self.data_base_url()}/model_aliases?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(list[ModelAlias], await resp.json())

    async def create_model_alias(self, model_alias: ModelAliasCreate, dry_run: bool = False) -> ModelAlias:
        post_url = f'{await self.data_base_url()}/model_aliases?account_uuid={self.account_uuid}&dry_run={dry_run}'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=model_alias.model_dump_json(exclude_unset=True)) as resp:
            return parse_obj_as(ModelAlias, await resp.json())

    async def get_model_alias(self, name: str) -> ModelAlias:
        get_url = f'{await self.data_base_url()}/model_aliases/{name}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(ModelAlias, await resp.json())

    async def delete_model_alias(self, name: str) -> None:
        delete_url = f'{await self.data_base_url()}/model_aliases/{name}'
        async with await self.request_with_retry("DELETE", delete_url):
            return

    async def update_model_alias(self, name: str, model_alias: ModelAliasUpdate) -> ModelAlias:
        patch_url = f'{await self.data_base_url()}/model_aliases/{name}'
        async with await self.request_with_retry("PATCH", patch_url, content_type=APPLICATION_JSON,
                                                 data=model_alias.model_dump_json(
                                                     exclude_unset=True, exclude_none=True
                                                 )) as resp:
            return parse_obj_as(ModelAlias, await resp.json())

    async def set_model_alias_tag(self, name: str, tag: str, model_uuid: str) -> None:
        patch_url = f'{await self.data_base_url()}/model_aliases/{name}/{tag}?model_uuid={model_uuid}'
        async with await self.request_with_retry("PATCH", patch_url):
            return

    async def delete_model_alias_tag(self, name: str, tag: str) -> None:
        delete_url = f'{await self.data_base_url()}/model_aliases/{name}/{tag}'
        async with await self.request_with_retry("DELETE", delete_url):
            return

    """ Arrow im and export methods """
    async def export_assets(
            self,
            dataset_uuid: str | None = None,
            dataset_version: int | None = None,
            asset_uuids: list[str] | None = None,
            model_uuid: str | None = None,
            transcode_mode: TranscodeMode = TranscodeMode.image_original_size,
            asset_url_type: AssetUrlType | None = None,
            inclusion_mode: AssetInclusionMode = AssetInclusionMode.annotated_only,
            annotation_inclusion_mode: AnnotationInclusionMode = AnnotationInclusionMode.all,
            include_external_ids: bool = False,
            freeze_dataset_version: bool | None = None,
            include_partitions: list[str] | None = None,
            include_auto_annotates: list[AutoAnnotate] | None = None,
            include_sources: list[str] | None = None,
    ) -> StreamReader:
        asset_url_type_query = f'asset_url_type={asset_url_type}&' if asset_url_type is not None else ''
        dataset_uuid_query = f'dataset_uuid={dataset_uuid}&' if dataset_uuid is not None else ''
        dataset_version_query = f'dataset_version={dataset_version}&' if dataset_version is not None else ''
        asset_uuids_query = ""
        if asset_uuids is not None:
            for asset_uuid in asset_uuids:
                asset_uuids_query += f"asset_uuid={asset_uuid}&"
        model_uuid_query = f'model_uuid={model_uuid}&' if model_uuid is not None else ''
        freeze_dataset_version_query = f'freeze_dataset_version={freeze_dataset_version}&' if freeze_dataset_version is not None else ''
        include_partitions_query = ""
        if include_partitions is not None:
            for include_partition in include_partitions:
                include_partitions_query += f"include_partition={include_partition}&"
        include_auto_annotates_query = ""
        if include_auto_annotates is not None:
            for include_auto_annotate in include_auto_annotates:
                include_auto_annotates_query += f"include_auto_annotate={include_auto_annotate}&"
        include_sources_query = ""
        if include_sources is not None:
            for include_source in include_sources:
                include_sources_query += f"include_source={include_source}&"
        get_url = (f'{await self.data_base_url()}/exports/assets?'
                   f'{dataset_uuid_query}'
                   f'{dataset_version_query}'
                   f'{asset_uuids_query}'
                   f'{model_uuid_query}'
                   f'transcode_mode={transcode_mode}&'
                   f'inclusion_mode={inclusion_mode}&'
                   f'annotation_inclusion_mode={annotation_inclusion_mode}&'
                   f'include_external_ids={"true" if include_external_ids else "false"}&'
                   f'{asset_url_type_query}'
                   f'{freeze_dataset_version_query}'
                   f'{include_partitions_query}'
                   f'{include_auto_annotates_query}'
                   f'{include_sources_query}')

        resp = await self.request_with_retry(
            method="GET",
            url=get_url,
            accept=MIME_TYPE_APACHE_ARROW_FILE_VERSIONED,
            timeout=aiohttp.ClientTimeout(total=None, sock_read=600)
        )
        return resp.content

    async def import_assets(
            self,
            arrow_stream: BinaryIO | AsyncIterable[bytes],
            dataset_uuid: str | None = None,
            dataset_version: int | None = None,
            model_uuid: str | None = None
    ) -> None:
        dataset_uuid_query = f'dataset_uuid={dataset_uuid}&' if dataset_uuid is not None else ''
        dataset_version_query = f'dataset_version={dataset_version}&' if dataset_version is not None else ''
        model_uuid_query = f'model_uuid={model_uuid}&' if model_uuid is not None else ''
        post_url = (f'{await self.data_base_url()}/imports/assets?'
                   f'{dataset_uuid_query}'
                   f'{dataset_version_query}'
                   f'{model_uuid_query}')
        await self.request_with_retry(
            "POST", post_url,
            data=arrow_stream,
            content_type=MIME_TYPE_APACHE_ARROW_FILE_VERSIONED,
            timeout=aiohttp.ClientTimeout(total=None, sock_read=600)
        )

    async def model_training_audits(self, model_uuids: list[str]) -> list[ModelTrainingAuditRecord]:
        model_uuid_query = ""
        for model_uuid in model_uuids:
            model_uuid_query += f"model_uuid={model_uuid}&"
        get_url = f'{await self.data_base_url()}/exports/model_training_audits?{model_uuid_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return TypeAdapter(list[ModelTrainingAuditRecord]).validate_python(await resp.json())


    async def export_model_urls(self, model_uuids: list[str], model_formats: list[ModelExportFormat], device_name: str | None) -> list[ExportedUrlResponse]:
        model_uuid_query = ""
        for model_uuid in model_uuids:
            model_uuid_query += f"model_uuid={model_uuid}&"
        model_format_query = ""
        for model_format in model_formats:
            model_format_query += f"model_format={model_format}&"
        device_name_query = f"device_name={quote_plus(device_name)}&" if device_name is not None else ""
        get_url = f'{await self.data_base_url()}/exports/model_urls?{model_uuid_query}{model_format_query}{device_name_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return TypeAdapter(list[ExportedUrlResponse]).validate_python(await resp.json())


    async def export_model_artifacts(self, model_uuids: list[str], model_formats: list[ModelExportFormat], device_name: str | None, artifact_type: ArtifactType | None) -> StreamReader:
        model_uuid_query = ""
        for model_uuid in model_uuids:
            model_uuid_query += f"model_uuid={model_uuid}&"
        model_format_query = ""
        for model_format in model_formats:
            model_format_query += f"model_format={model_format}&"
        device_name_query = f"device_name={quote_plus(device_name)}&" if device_name is not None else ""
        type_query = f"artifact_type={artifact_type}&" if artifact_type is not None else ""

        get_url = f'{await self.data_base_url()}/exports/model_artifacts?{model_uuid_query}{model_format_query}{device_name_query}{type_query}'
        resp = await self.request_with_retry("GET", get_url)
        return resp.content


    async def model_training_event(self, model_training_event: ModelTrainingEvent, model_uuid: str) -> None:
        post_url = f'{await self.data_base_url()}/models/{model_uuid}/events'
        await self.request_with_retry(
            "POST", post_url,
            data=model_training_event.model_dump_json(exclude_unset=True),
            content_type=APPLICATION_JSON,
        )

    async def qc_ai_hub_model_export(self, model_uuid: str, export_params: list[QcAiHubExportParams]) -> None:
        post_url = f'{await self.data_base_url()}/models/{model_uuid}/exports/qc_ai_hub'
        await self.request_with_retry(
            "POST", post_url,
            data=TypeAdapter(list[QcAiHubExportParams]).dump_json(export_params),
            content_type=APPLICATION_JSON,
        )

    # --- Workflow Endpoints ---
    async def start_workflow(
        self,
        template_name: str,
        workflow_create: CreateWorkflowBody,
    ) -> CreateWorkflowResponse:
        query = f'template_name={template_name}'
        query += f'&account_uuid={self.account_uuid}'
        post_url = f'{await self.data_base_url()}/workflows?{query}'
        async with await self.request_with_retry(
            "POST", post_url,
            content_type=APPLICATION_JSON,
            data=workflow_create.model_dump_json(exclude_unset=True)
        ) as resp:
            return parse_obj_as(CreateWorkflowResponse, await resp.json())

    async def get_workflow(self, workflow_id: str) -> ListWorkflowItem:
        get_url = f'{await self.data_base_url()}/workflows/{workflow_id}?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(ListWorkflowItem, await resp.json())

    async def list_workflows(
        self,
        dataset_uuid: list[str] | None = None,
        model_uuid: list[str] | None = None,
        phase: list[ArgoWorkflowPhase] | None = None
    ) -> list[ListWorkflowItem]:
        query = f'account_uuid={self.account_uuid}'
        if dataset_uuid:
            for uuid in dataset_uuid:
                query += f'&dataset_uuid={uuid}'
        if model_uuid:
            for uuid in model_uuid:
                query += f'&model_uuid={uuid}'
        if phase:
            for p in phase:
                query += f'&phase={p.value}'
        get_url = f'{await self.data_base_url()}/workflows?{query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(list[ListWorkflowItem], await resp.json())



import logging
import time
from enum import Enum

from eyepop.jobs import JobState
from eyepop.worker.worker_jobs import JobStateCallback, WorkerJob

log = logging.getLogger('eyepop')

class MetricCollector(JobStateCallback):
    def __init__(self):
        self.jobs_to_state = {}
        self.jobs_to_last_updated = {}
        self.total_number_of_jobs = 0
        self.max_number_of_jobs_by_state = {JobState.STARTED: 0, JobState.IN_PROGRESS: 0, JobState.FINISHED: 0,
                                            JobState.DRAINED: 0, JobState.FAILED: 0, }
        self.number_of_jobs_reached_state = {JobState.IN_PROGRESS: 0, JobState.FINISHED: 0, JobState.DRAINED: 0,
                                             JobState.FAILED: 0, }
        self.total_time_to_reached_state = {JobState.IN_PROGRESS: 0.0, JobState.FINISHED: 0.0, JobState.DRAINED: 0.0,
                                            JobState.FAILED: 0.0, }

    def get_average_time(self, state: JobState) -> float:
        if self.number_of_jobs_reached_state[state] == 0:
            return 0.0
        else:
            return self.total_time_to_reached_state[state] / self.number_of_jobs_reached_state[state]

    def get_average_times(self) -> dict:
        times = {JobState.IN_PROGRESS: self.get_average_time(JobState.IN_PROGRESS),
                 JobState.FINISHED: self.get_average_time(JobState.FINISHED),
                 JobState.DRAINED: self.get_average_time(JobState.DRAINED),
                 JobState.FAILED: self.get_average_time(JobState.FAILED), }
        return times

    def update_count_by_state(self, state: JobState):
        count = 0
        for job, job_state in self.jobs_to_state.items():
            if state == job_state:
                count += 1
        if count > self.max_number_of_jobs_by_state[state]:
            self.max_number_of_jobs_by_state[state] = count

    def collect_execution_time(self, job: WorkerJob, new_state: JobState):
        now = time.time()
        duration = now - self.jobs_to_last_updated[job]
        self.jobs_to_last_updated[job] = now
        self.number_of_jobs_reached_state[new_state] += 1
        self.total_time_to_reached_state[new_state] += duration

    def created(self, job):
        self.total_number_of_jobs += 1
        self.jobs_to_state[job] = JobState.CREATED
        self.jobs_to_last_updated[job] = time.time()

    def started(self, job):
        current_state = self.jobs_to_state[job]
        if current_state != JobState.CREATED:
            log.debug("invalid job state %v in metrics collector for started(%v)", current_state, job)
        else:
            self.jobs_to_state[job] = JobState.STARTED
        self.jobs_to_last_updated[job] = time.time()
        self.update_count_by_state(JobState.STARTED)

    def first_result(self, job):
        current_state = self.jobs_to_state[job]
        if current_state != JobState.STARTED:
            log.debug("invalid job state %v in metrics collector for first_result(%v)", current_state, job)
        else:
            self.jobs_to_state[job] = JobState.IN_PROGRESS
        self.collect_execution_time(job, JobState.IN_PROGRESS)
        self.update_count_by_state(JobState.IN_PROGRESS)

    def failed(self, job):
        self.jobs_to_state[job] = JobState.FAILED
        self.collect_execution_time(job, JobState.FAILED)
        self.update_count_by_state(JobState.FAILED)
        self.jobs_to_last_updated[job] = time.time()

    def finished(self, job):
        self.jobs_to_state[job] = JobState.FINISHED
        self.collect_execution_time(job, JobState.FINISHED)
        self.update_count_by_state(JobState.FINISHED)
        self.jobs_to_last_updated[job] = time.time()

    def drained(self, job):
        self.jobs_to_state[job] = JobState.DRAINED
        self.collect_execution_time(job, JobState.DRAINED)
        self.update_count_by_state(JobState.DRAINED)
        self.jobs_to_last_updated[job] = time.time()

    def finalized(self, job):
        del self.jobs_to_state[job]


class JobState(Enum):
    CREATED = 1
    STARTED = 2
    IN_PROGRESS = 3
    FINISHED = 4
    FAILED = 5
    DRAINED = 6

    def __repr__(self):
        return self._name_



import enum
from typing import List, Literal, Annotated, Union, Any

from pydantic import BaseModel, Field, ConfigDict

class PredictionVersion(enum.IntEnum):
    V1 = 1
    V2 = 2


DEFAULT_PREDICTION_VERSION = PredictionVersion.V2


class VideoMode(enum.StrEnum):
    STREAM = "stream"
    BUFFER = "buffer"


class PopComponentType(enum.StrEnum):
    BASE = "<invalid>"
    FORWARD = "forward"
    INFERENCE = "inference"
    # backward compatibility for persisted Pops < 3.0.0
    TRACING = "tracing"
    # since 3.0.0, replaced 'tracing'
    TRACKING = "tracking"
    CONTOUR_FINDER = "contour_finder"
    COMPONENT_FINDER = "component_finder"


class ForwardOperatorType(enum.StrEnum):
    FULL = "full"
    CROP = "crop"
    CROP_WITH_FULL_FALLBACK = "crop_with_full_fallback"


class PopCrop(BaseModel):
    maxItems: int | None = None
    boxPadding: float | None = None
    orientationTargetAngle: float | None = None
    model_config = ConfigDict(extra='forbid')


class PopForwardOperator(BaseModel):
    type: ForwardOperatorType
    includeClasses: list[str] | None = None
    crop: PopCrop | None = None
    model_config = ConfigDict(extra='forbid')

class PopForward(BaseModel):
    operator: PopForwardOperator | None = None
    targets: List["DynamicComponent"] | None = None
    model_config = ConfigDict(extra='forbid')

class BaseComponent(BaseModel):
    type: Literal[PopComponentType.BASE] = PopComponentType.BASE
    id: int | None = None
    forward: PopForward | None = None
    model_config = ConfigDict(extra='forbid')


class ForwardComponent(BaseComponent):
    type: Literal[PopComponentType.FORWARD] = PopComponentType.FORWARD
    model_config = ConfigDict(extra='forbid')


class InferenceType(enum.StrEnum):
    IMAGE_CLASSIFICATION = "image_classification"
    OBJECT_DETECTION = "object_detection"
    KEY_POINTS = "key_points"
    OCR = "ocr"
    MESH = "mesh"
    FEATURE_VECTOR = "feature_vector"
    SEMANTIC_SEGMENTATION = "semantic_segmentation"
    SEGMENTATION = "segmentation"
    RAW = "raw"


class InferenceComponent(BaseComponent):
    type: Literal[PopComponentType.INFERENCE] = PopComponentType.INFERENCE
    inferenceTypes: List[InferenceType] | None = None
    hidden: bool | None = None
    modelUuid: Annotated[str | None, Field(default=None, deprecated='modelUuid is deprecated, use abilityUuid instead'), ]
    model: Annotated[str | None, Field(default=None, deprecated='model is deprecated, use ability instead'), ]
    abilityUuid: str | None = None
    ability: str | None = None
    categoryName: str | None = None
    confidenceThreshold: float | None = None
    topK: int | None = None
    targetFps: str | None = None
    params: dict[str, Any] | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')


class TrackingComponent(BaseComponent):
    type: Literal[PopComponentType.TRACKING] = PopComponentType.TRACKING
    reidModelUuid: str | None = None
    reidModel: str | None = None
    maxAgeSeconds: float | None = None
    iouThreshold: float | None = None
    simThreshold: float | None = None
    agnostic: bool | None = None
    model_config = ConfigDict(extra='forbid')


class ContourType(enum.StrEnum):
    ALL_PIXELS = "all_pixels"
    POLYGON = "polygon"
    CONVEX_HULL = "convex_hull"
    HOUGH_CIRCLES = "hough_circles"
    CIRCLE = "circle"
    TRIANGLE = "triangle"
    RECTANGLE = "rectangle"


class ContourFinderComponent(BaseComponent):
    type: Literal[PopComponentType.CONTOUR_FINDER] = PopComponentType.CONTOUR_FINDER
    contourType: ContourType
    areaThreshold: float | None = None
    model_config = ConfigDict(extra='forbid')


class ComponentFinderComponent(BaseComponent):
    type: Literal[PopComponentType.COMPONENT_FINDER] = PopComponentType.COMPONENT_FINDER
    dilate: float | None = None
    erode: float | None = None
    keepSource: bool | None = None
    componentClassLabel: str | None = None
    model_config = ConfigDict(extra='forbid')


DynamicComponent = Annotated[Union[ForwardComponent | InferenceComponent | TrackingComponent | ContourFinderComponent | ComponentFinderComponent], Field(discriminator="type")]


class Pop(BaseModel):
    components: List[DynamicComponent]
    postTransform: str | None = None
    model_config = ConfigDict(extra='forbid')

# Helper factories

def CropForward(
        targets: List[DynamicComponent],
        maxItems: int | None = None,
        boxPadding: float | None = None,
        orientationTargetAngle: float | None = None,
        includeClasses: list[str] | None = None,
        is_full_fallback: bool = False
) -> PopForward:
    return PopForward(
        operator=PopForwardOperator(
            type=ForwardOperatorType.CROP if not is_full_fallback else ForwardOperatorType.CROP_WITH_FULL_FALLBACK,
            includeClasses=includeClasses,
            crop=PopCrop(
                maxItems=maxItems,
                boxPadding=boxPadding,
                orientationTargetAngle=orientationTargetAngle,
            ),
        ),
        targets=targets
    )

def FullForward(
        targets: List[DynamicComponent],
        includeClasses: list[str] | None = None
) -> PopForward:
    return PopForward(
        operator=PopForwardOperator(
            type=ForwardOperatorType.FULL,
            includeClasses=includeClasses
        ),
        targets=targets,
    )


class ComponentParams(BaseComponent):
    componentId: int
    values: dict[str, Any]
    model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')
